""" AppOptics APM instrumentation for memcache client module.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""
from appoptics_apm import util
import socket
from functools import partial, wraps

appoptics_apm_logger = util.logger

# memcache.Client methods (from docstring)
# Setup: __init__, set_servers, forget_dead_hosts, disconnect_all, debuglog
# Insertion: set, add, replace, set_multi
# Retrieval: get, get_multi
# Integers: incr, decr
# Removal: delete, delete_multi
# Mutate: append, cas, prepend

# memcache.Client setup
MC_SERVER_COMMANDS = set(('__init__', 'set_servers'))

# these methods also have the same names as Memcached commands/ops
MC_COMMANDS = set(('get', 'get_multi',
                   'set', 'add', 'replace', 'set_multi',
                   'incr', 'decr',
                   'delete', 'delete_multi',
                   'append', 'cas', 'prepend', 'gets'))


def wrap_mc_method(func, f_args, f_kwargs, return_val, funcname=None):
    """Pulls the operation and (for get) whether a key was found, on each public method."""
    kvs = {}
    if funcname in MC_COMMANDS:
        kvs['KVOp'] = funcname
    # could examine f_args for key(s) here
    if funcname == 'get':
        kvs['KVHit'] = int(return_val is not None)
    return kvs


def wrap_get_server(layer_name, func):
    """ Wrapper for memcache._get_server, to read remote host on all ops.

    This relies on the module internals, and just sends an info event when this
    function is called.
    """
    @wraps(func) # XXX Not Python2.4-friendly
    def wrapper(*f_args, **f_kwargs):
        ret = func(*f_args, **f_kwargs)
        try:
            args = {'KVKey': f_args[1]}
            (host, _) = ret
            if host:
                if host.family == socket.AF_INET:
                    args['RemoteHost'] = host.ip
                elif host.family == socket.AF_UNIX:
                    args['RemoteHost'] = 'localhost'

            util.log('info', layer_name, keys=args,
                     store_backtrace=util._collect_backtraces('memcache')
                     )
        except Exception as e:
            appoptics_apm_logger.error("AppOptics APM error: %s" % str(e))

        return ret
    return wrapper


def dynamic_wrap(fn):
    # We explicity pass assigned to wraps; this skips __module__ from the
    # default list, which doesn't exist for the functions from pylibmc.
    @wraps(fn, assigned=('__name__', '__doc__'))
    def wrapped(*args, **kwargs):
        return fn(*args, **kwargs)
    return wrapped


def wrap(layer_name, module):
    try:
        # wrap middleware callables we want to wrap
        cls = getattr(module, 'Client', None)
        if not cls:
            return
        for method in MC_COMMANDS:
            # delete_multi delegates to delete in pylibmc, so don't instrument it
            if method == 'delete_multi' and module.__name__ == 'pylibmc':
                continue
            fn = getattr(cls, method, None)
            if not fn:
                # this method does not exist for this module/version
                continue
            kvs = {'Class': layer_name + '.Client',
                   'Function': method}
            wrapfn = fn if hasattr(fn, '__func__') else dynamic_wrap(fn)
            wrapper = util.log_method(layer_name,
                                      callback=partial(wrap_mc_method, funcname=method),
                                      entry_kvs=kvs)
            setattr(cls, method, wrapper(wrapfn))

        # per-key memcache host hook
        if hasattr(cls, '_get_server'):
            fn = getattr(cls, '_get_server', None)
            setattr(cls, '_get_server', wrap_get_server(layer_name, fn))

    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM error: %s" % str(e))

if util.ready():
    for module_name in ['memcache', 'pylibmc']:
        try:
            mod = __import__(module_name)
            wrap(module_name, mod)
        except (ImportError, KeyError) as ex:
            pass



# Modified from https://github.com/dcrosta/professor/blob/master/professor/skeleton.py
#
# Copyright (c) 2011, Daniel Crosta
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import six
from bson.objectid import ObjectId
from bson.code import Code
from bson.son import SON
from bson.dbref import DBRef
from datetime import datetime
import re
from appoptics_apm.util import logger

__all__ = ('skeleton', 'sanitize')

BSON_TYPES = (set(six.integer_types) | set(six.string_types)
              | set([bool, float, datetime, ObjectId, type(re.compile('')), Code, type(None), DBRef, SON, ]))


def skeleton(query_part, preserve_first=False, keep_values=False, no_wrap=False):
    # >>> skeleton({u'acronym': u'AMSHAT'})
    # u'{acronym}'
    # >>> skeleton({u'_id': 1, u'locked': False})
    # u'{_id,locked}'
    # >>> skeleton({})
    # u'{}'
    # >>> skeleton({u'op': u'query'})
    # u'{op}'
    t = type(query_part)

    if t == SON:
        query_part = query_part.to_dict()
        t = type(query_part)
    elif query_part is None:
        query_part = ''
        t = type(query_part)

    if t == list:
        out = []
        for element in query_part:
            if preserve_first:
                sub = skeleton(element, keep_values=True)
                preserve_first = False
            else:
                sub = skeleton(element)
            if sub is not None:
                out.append(sub)
        if no_wrap:
            return six.u(','.join(out))
        else:
            return six.u('[{o}]'.format(o=','.join(out)))
    elif t == dict:
        out = []
        for key in sorted(query_part.keys()):
            sub = skeleton(query_part[key], keep_values=keep_values)
            if sub is not None:
                out.append('%s:%s' % (key, sub))
            else:
                out.append(key)
        if no_wrap:
            return six.u(','.join(out))
        else:
            return six.u('{%s}' % ','.join(out))
    elif t not in BSON_TYPES:
        logger.debug('Value: {} is of an unsupported type.'.format(query_part))
        return
    elif keep_values:
        return str(query_part)


def sanitize(value):
    # return a copy of the query with all
    # occurrences of "$" replaced by "_$_",
    # and ocurrences of "." replaced by
    # "_,_" in keys
    t = type(value)
    if t == list:
        return map(sanitize, value)
    elif t == dict:
        return dict((k.replace('$', '_$_').replace('.', '_,_'), sanitize(v))
                    for k, v in value.items())
    elif t not in BSON_TYPES:
        logger.debug('Value: {} is of an unsupported type.'.format(value))
        return
    else:
        return value


def desanitize(value):
    # perform the inverse of sanitize()
    t = type(value)
    if t == list:
        return map(desanitize, value)
    elif t == dict:
        return dict((k.replace('_$_', '$').replace('_,_', '.'), desanitize(v))
                    for k, v in value.items())
    elif t not in BSON_TYPES:
        logger.debug('Value: {} is of an unsupported type.'.format(value))
        return
    else:
        return value



""" Handles JSON conversion for various object types that may be found in queries

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""
try:
    import json
except ImportError:
    import simplejson as json

import datetime
from bson.objectid import ObjectId
from bson.dbref import DBRef


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ObjectId) or isinstance(obj, DBRef):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%dT%H:%M:%S')
        else:
            try:
                return json.JSONEncoder.default(self, obj)
            except Exception:
                # so we don't fail the request:
                return "[unsupported type %s]" % (type(obj))



# Copyright (C) 1996-2004 by Phillip J. Eby and Tyler C. Sarna.
# All rights reserved.
#
# This license has been certified as open source. It has also been designated as
# GPL compatible by the Free Software Foundation (FSF).
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions in source code must retain the accompanying copyright
#      notice, this list of conditions, and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the accompanying copyright
#      notice, this list of conditions, and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Names of the copyright holders must not be used to endorse or promote
#      products derived from this software without prior written permission
#      from the copyright holders.
#
#   4. The right to distribute this software or to use it for any purpose does
#      not give you the right to use Servicemarks (sm) or Trademarks (tm) of
#      the copyright holders. Use of them is covered by separate agreement with
#      the copyright holders.
#
#   5. If any files are modified, you must cause the modified files to carry
#      prominent notices stating that you changed the files and the date of any
#      change.
#
#
# Disclaimer
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY EXPRESSED
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tools for doing dynamic imports"""

__all__ = [
    'importString', 'importObject', 'importSequence', 'importSuite',
    'lazyModule', 'joinPath', 'whenImported', 'getModuleHooks',
]

import sys
import six
from types import ModuleType
from imp import acquire_lock, release_lock

try:
    import __main__
except ImportError:
    pass

defaultGlobalDict = __main__.__dict__

try:
    from peak.util.EigenData import AlreadyRead
except ImportError:
    class AlreadyRead(Exception):pass


def importSuite(specs, globalDict=defaultGlobalDict):
    """Create a test suite from import specs"""
    from unittest import TestSuite
    return TestSuite(
        [t() for t in importSequence(specs,globalDict)]
    )


def joinPath(modname, relativePath):
    """Adjust a module name by a '/'-separated, relative or absolute path"""
    module = modname.split('.')
    for p in relativePath.split('/'):

        if p=='..':
            module.pop()
        elif not p:
            module = []
        elif p!='.':
            module.append(p)

    return '.'.join(module)


def importString(name, globalDict=defaultGlobalDict):
    """Import an item specified by a string

    Example Usage::

        attribute1 = importString('some.module:attribute1')
        attribute2 = importString('other.module:nested.attribute2')

    'importString' imports an object from a module, according to an
    import specification string: a dot-delimited path to an object
    in the Python package namespace.  For example, the string
    '"some.module.attribute"' is equivalent to the result of
    'from some.module import attribute'.

    For readability of import strings, it's sometimes helpful to use a ':' to
    separate a module name from items it contains.  It's optional, though,
    as 'importString' will convert the ':' to a '.' internally anyway."""

    if ':' in name:
        name = name.replace(':','.')

    parts = filter(None,name.split('.'))
    item = __import__(parts.pop(0), globalDict, globalDict, ['__name__'])

    # Fast path for the common case, where everything is imported already
    for attr in parts:
        try:
            item = getattr(item, attr)
        except AttributeError:
            break   # either there's an error, or something needs importing
    else:
        return item

    # We couldn't get there with just getattrs from the base import.  So now
    # we loop *backwards* trying to import longer names, then shorter, until
    # we find the longest possible name that can be handled with __import__,
    # then loop forward again with getattr.  This lets us give more meaningful
    # error messages than if we only went forwards.
    attrs = []
    exc = None

    try:
        while True:
            try:
                # Exit as soon as we find a prefix of the original `name`
                # that's an importable *module* or package
                item = __import__(name, globalDict, globalDict, ['__name__'])
                break
            except ImportError:
                if not exc:
                    # Save the first ImportError, as it's usually the most
                    # informative, especially w/Python < 2.4
                    exc = sys.exc_info()

                if '.' not in name:
                    # We've backed up all the way to the beginning, so reraise
                    # the first ImportError we got
                    six.reraise(*exc)

                # Otherwise back up one position and try again
                parts = name.split('.')
                attrs.append(parts[-1])
                name = '.'.join(parts[:-1])
    finally:
        exc = None

    # Okay, the module object is now in 'item', so we can just loop forward
    # to retrieving the desired attribute.
    #
    while attrs:
        attr = attrs.pop()
        try:
            item = getattr(item,attr)
        except AttributeError:
            raise ImportError("%r has no %r attribute" % (item,attr))

    return item


def lazyModule(modname, relativePath=None):
    """Return module 'modname', but with its contents loaded "on demand"

    This function returns 'sys.modules[modname]', if present.  Otherwise
    it creates a 'LazyModule' object for the specified module, caches it
    in 'sys.modules', and returns it.

    'LazyModule' is a subclass of the standard Python module type, that
    remains empty until an attempt is made to access one of its
    attributes.  At that moment, the module is loaded into memory, and
    any hooks that were defined via 'whenImported()' are invoked.

    Note that calling 'lazyModule' with the name of a non-existent or
    unimportable module will delay the 'ImportError' until the moment
    access is attempted.  The 'ImportError' will occur every time an
    attribute access is attempted, until the problem is corrected.

    This function also takes an optional second parameter, 'relativePath',
    which will be interpreted as a '/'-separated path string relative to
    'modname'.  If a 'relativePath' is supplied, the module found by
    traversing the path will be loaded instead of 'modname'.  In the path,
    '.' refers to the current module, and '..' to the current module's
    parent.  For example::

        fooBaz = lazyModule('foo.bar','../baz')

    will return the module 'foo.baz'.  The main use of the 'relativePath'
    feature is to allow relative imports in modules that are intended for
    use with module inheritance.  Where an absolute import would be carried
    over as-is into the inheriting module, an import relative to '__name__'
    will be relative to the inheriting module, e.g.::

        something = lazyModule(__name__,'../path/to/something')

    The above code will have different results in each module that inherits
    it.

    (Note: 'relativePath' can also be an absolute path (starting with '/');
    this is mainly useful for module '__bases__' lists.)"""

    def _loadModule(module):
        oldGA = LazyModule.__getattribute__
        oldSA = LazyModule.__setattr__

        modGA = ModuleType.__getattribute__
        modSA = ModuleType.__setattr__

        LazyModule.__getattribute__ = modGA
        LazyModule.__setattr__      = modSA

        acquire_lock()
        try:
            try:
                # don't reload if already loaded!
                if set(module.__dict__.keys()) == set(['__name__']):
                    # Get Python to do the real import!
                    if six.PY2:
                        module.__name__ = module.__name__.encode('utf-8')
                    six.moves.reload_module(module)
                    
                try:
                    for hook in getModuleHooks(module.__name__):
                        hook(module)
                finally:
                    # Ensure hooks are not called again, even if they fail
                    postLoadHooks[module.__name__] = None
            except Exception:
                # Reset our state so that we can retry later
                if '__file__' not in module.__dict__:
                    LazyModule.__getattribute__ = oldGA.__func__
                    LazyModule.__setattr__ = oldSA.__func__
                raise

            try:
                # Convert to a real module (if under 2.2)
                module.__class__ = ModuleType
            except TypeError:
                pass    # 2.3 will fail, but no big deal

        finally:
            release_lock()

    class LazyModule(ModuleType):
        __slots__ = ()
        def __init__(self, name):
            ModuleType.__setattr__(self,'__name__',name)
            # super(LazyModule,self).__init__(name)

        def __getattribute__(self,attr):
            _loadModule(self)
            return ModuleType.__getattribute__(self,attr)

        def __setattr__(self,attr,value):
            _loadModule(self)
            return ModuleType.__setattr__(self,attr,value)

    if relativePath:
        modname = joinPath(modname, relativePath)
    acquire_lock()
    try:
        if modname not in sys.modules:
            getModuleHooks(modname) # force an empty hook list into existence
            sys.modules[modname] = LazyModule(modname)
            if '.' in modname:
                # ensure parent module/package is in sys.modules
                # and parent.modname=module, as soon as the parent is imported
                splitpos = modname.rindex('.')
                whenImported(
                    modname[:splitpos],
                    lambda m: setattr(m,modname[splitpos+1:], sys.modules[modname])
                )
        return sys.modules[modname]
    finally:
        release_lock()


postLoadHooks = {}


def getModuleHooks(moduleName):
    """Get list of hooks for 'moduleName'; error if module already loaded"""
    acquire_lock()
    try:
        hooks = postLoadHooks.setdefault(moduleName, [])
        if hooks is None:
            raise AlreadyRead("Module already imported", moduleName)
        return hooks
    finally:
        release_lock()


def _setModuleHook(moduleName, hook):
    acquire_lock()
    # Lazy loading does not work on py3.3 (it uses type(module) == ModuleType for the parameter check.
    if moduleName not in sys.modules and sys.version_info[0] == 3 and sys.version_info[1] == 3:
        import importlib
        try:
            importlib.import_module(moduleName)
        except ImportError:
            pass

    try:
        if moduleName in sys.modules and postLoadHooks.get(moduleName) is None:
            # Module is already imported/loaded, just call the hook
            module = sys.modules[moduleName]
            hook(module)
            return module
        
        # Lazy loading does not work on py3.3
        if not(sys.version_info[0] == 3 and sys.version_info[1] == 3):
            getModuleHooks(moduleName).append(hook)
            return lazyModule(moduleName)
    finally:
        release_lock()


def whenImported(moduleName, hook=None):
    """Call 'hook(module)' when module named 'moduleName' is first used

    'hook' must accept one argument: the module object named by 'moduleName',
    which must be a fully qualified (i.e. absolute) module name.  The hook
    should not raise any exceptions, or it may prevent later hooks from
    running.

    If the module has already been imported normally, 'hook(module)' is
    called immediately, and the module object is returned from this function.
    If the module has not been imported, or has only been imported lazily,
    then the hook is called when the module is first used, and a lazy import
    of the module is returned from this function.  If the module was imported
    lazily and used before calling this function, the hook is called
    immediately, and the loaded module is returned from this function.

    Note that using this function implies a possible lazy import of the
    specified module, and lazy importing means that any 'ImportError' will be
    deferred until the module is used.
    """
    if hook is None:
        def decorate(func):
            whenImported(moduleName, func)
            return func
        return decorate

    if '.' in moduleName:
        # If parent is not yet imported, delay hook installation until the
        # parent is imported.
        splitpos = moduleName.rindex('.')
        whenImported(
            moduleName[:splitpos], lambda m: _setModuleHook(moduleName, hook)
        )
    else:
        return _setModuleHook(moduleName, hook)

def importObject(spec, globalDict=defaultGlobalDict):
    """Convert a possible string specifier to an object

    If 'spec' is a string or unicode object, import it using 'importString()',
    otherwise return it as-is.
    """

    if isinstance(spec, six.string_types):
        return importString(spec, globalDict)

    return spec


def importSequence(specs, globalDict=defaultGlobalDict):
    """Convert a string or list specifier to a list of objects.

    If 'specs' is a string or unicode object, treat it as a
    comma-separated list of import specifications, and return a
    list of the imported objects.

    If the result is not a string but is iterable, return a list
    with any string/unicode items replaced with their corresponding
    imports.
    """
    if isinstance(specs, six.string_types):
        return [importString(x.strip(),globalDict) for x in specs.split(',')]
    else:
        return [importObject(s,globalDict) for s in specs]




""" AppOptics APM instrumentation for pymongo (MongoDB client).

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""
import six
from appoptics_apm import util

try:
    import json
except ImportError:
    import simplejson as json

PYMONGO_LAYER = "pymongo"
appoptics_apm_logger = util.logger


# Helper methods for pymongo instrumentation:
def profile_find(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo 'find' profiling """
    collection = args[0]
    if len(args) > 1:
        query = args[1]
    else:
        query = 'all'

    return _profile_query(collection, query, op='find')


def profile_update(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo 'update' profiling """
    (collection, query, doc, _upsert, _manipulate, safe, multi) = args[:7]

    report_kvs = _profile_query(
        collection, query, op='update', safe=safe, result=func_result
    )
    report_kvs['Update_Document'] = _to_json(doc)

    if multi:
        report_kvs['Multi'] = True

    return report_kvs


def profile_insert(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo 'insert' profiling """
    (collection, docs, _manipulate, safe) = args[:4]
    if isinstance(docs, dict):
        docs = [docs]

    # Only first doc is finger printed
    report_kvs = _profile_query(
        collection, docs[0], op='insert', safe=safe,
        result=func_result, docs_affected=len(docs)
    )

    return report_kvs


def profile_remove(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo 'remove' profiling """
    (collection, spec_or_id, safe) = args[:3]

    if spec_or_id is None:
        spec_or_id = {}

    if not isinstance(spec_or_id, dict):
        spec_or_id = {"_id": spec_or_id}

    return _profile_query(
        collection, spec_or_id, op='remove', safe=safe, result=func_result
    )


def profile_drop(func, args, kwargs, func_results):
    """ AppOptics APM callback for pymongo 'drop' profiling """
    return _profile_query(args[0], None)


def profile_index(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo create/remove/ensure index profiling """
    collection = args[0]

    report_kvs = _profile_query(collection)

    if len(args) > 1:
        report_kvs['Index'] = _to_json(args[1])

    return report_kvs


def profile_group(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo collection group profiling """
    (collection, key, condition, initial, reduce) = args[:5]
    report_kvs = _profile_query(collection)

    if key:
        report_kvs['Group_Key'] = _to_json(key)

    if condition:
        report_kvs['Group_Condition'] = _to_json(condition)

    if initial:
        report_kvs['Group_Initial'] = _to_json(initial)

    if reduce:
        report_kvs['Group_Reduce'] = reduce

    return report_kvs


def profile_distinct(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo collection distinct profiling """
    collection = args[0]

    report_kvs = _profile_query(collection)
    report_kvs['Key'] = args[1]

    return report_kvs


def profile_map_reduce(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo collection map/reduce """
    (collection, map_fn, reduce_fn) = args[:3]

    report_kvs = _profile_query(collection)
    report_kvs['Map_Function'] = map_fn
    report_kvs['Reduce_Function'] = reduce_fn

    return report_kvs


def profile_find_and_modify(func, args, kwargs, func_result):
    """ AppOptics APM callback for pymongo collection map/reduce """
    (collection, query, update) = args[:3]

    report_kvs = _profile_query(collection, query, op='find_and_modify')
    if update is not None:
        report_kvs['Update_Document'] = _to_json(update)

    return report_kvs


def profile_create_collection(func, args, kwargs, func_results):
    """ AppOptics APM callback for pymongo 'create collection' profiling """
    collection = func_results
    if collection is not None:
        return _profile_query(collection)


def profile_rename_collection(func, args, kwargs, func_results):
    """ AppOptics APM callback for pymongo 'rename collection' profiling """
    (collection, new_name) = args[:2]
    report_kvs = _profile_query(collection)
    report_kvs['New_Collection_Name'] = new_name
    return report_kvs


def profile_drop_collection(func, args, kwargs, func_results):
    """ AppOptics APM callback for pymongo 'drop collection' profiling """
    (db, name_or_collection) = args[:2]
    report_kvs = {}

    if isinstance(name_or_collection, six.string_types):
        collection_name = name_or_collection
    else:
        collection_name = name_or_collection.name

    _add_connection_info(report_kvs, db)
    report_kvs['Collection'] = collection_name

    return report_kvs


def profile_command(func, args, kwargs, func_results):
    """ AppOptics APM callback for pymongo database 'command' profiling """
    (db, command) = args[:2]
    report_kvs = {}

    _add_connection_info(report_kvs, db)
    if not isinstance(command, six.string_types):
        command = _to_json(command)

    report_kvs['Command'] = command

    return report_kvs


def _profile_query(collection, query=None, op=None, safe=None, result=None, docs_affected=None):
    """ Gathers report key/values from collection and (optional) query """

    report_kvs = {}
    _add_connection_info(report_kvs, collection.database)
    report_kvs['Collection'] = collection.name

    if isinstance(query, dict):
        report_kvs['Query'] = _to_json(query)
        if op:
            if op == 'find' and collection.name == '$cmd':
                report_kvs['QueryFingerprint'] = '%s.command(%s)' % (
                    collection.database.name, _command_fingerprint(query))
            else:
                report_kvs['QueryFingerprint'] = '%s.%s.%s(%s)' % (
                    collection.database.name, collection.name, op, _query_fingerprint(query))

    if op:
        report_kvs['QueryOp'] = op

    if safe is not None:
        report_kvs['SafeMode'] = safe
        # We only get document count if safe mode is true
        if safe and docs_affected is None and result is not None and 'n' in result:
            report_kvs['NumDocumentsAffected'] = result['n']

    if docs_affected is not None:  # Used for insert
        report_kvs['NumDocumentsAffected'] = docs_affected

    return report_kvs


def _add_connection_info(report_kvs, db):
    """ Connection info common to every mongodb trace """
    report_kvs['Flavor'] = 'mongodb'
    report_kvs['Database'] = db.name
    # The MongoDB host may be a sharded cluster which contains multiple hosts
    report_kvs['RemoteHostPool'] = str(list(db.client.nodes))


def _to_json(obj):
    try:
        return json.dumps(obj, cls=JSONEncoder)
    except Exception as e:
        return "{'AppOpticsApmError': 'Could not dump data to JSON.'}"


def _query_fingerprint(query):
    try:
        return skeleton(dict(query))
    except:
        # For safety, in case skeleton encounters something unexpected we won't fail the query:
        return ""


def _command_fingerprint(query):
    """ Special case for commands: preserve order of arguments, and save
        first argument name/value
    """
    fp = None
    if query and isinstance(query, SON) and len(query) > 0:
        cmd_args = [{item[0]: item[1]} for item in query.items()]
        try:
            fp = skeleton(cmd_args, preserve_first=True, no_wrap=True)
        except:
            # For safety, in case skeleton encounters something unexpected we won't fail the query:
            pass

    return fp


# Parameter dicts for method instrumentation:

# For pymongo.collection.Collection:
COLLECTION_METHOD_INST = {
    'save': {},
    'insert': {
        'callback': profile_insert,
        'store_backtrace': True,
    },
    'update': {
        'callback': profile_update,
        'store_backtrace': True,
    },
    'drop': {
        'callback': profile_drop,
        'store_backtrace': True,
    },
    'remove': {
        'callback': profile_remove,
        'store_backtrace': True,
    },
    'find': {
        'callback': profile_find,
        'store_backtrace': True,
    },
    'count': {},
    'create_index': {
        'callback': profile_index,
        'store_backtrace': True,
    },
    'ensure_index': {
        'callback': profile_index,
        'store_backtrace': True,
    },
    'drop_indexes': {
        'callback': profile_index,
    },
    'drop_index': {
        'callback': profile_index,
        'store_backtrace': True,
    },
    'reindex': {
        'callback': profile_index,
        'store_backtrace': True,
    },
    'index_information': {
        'callback': profile_index,
        'store_backtrace': True,
    },
    'options': {
        'store_backtrace': True,
    },
    'group': {
        'callback': profile_group,
        'store_backtrace': True,
    },
    'rename': {
        'callback': profile_rename_collection,
        'store_backtrace': True,
    },
    'distinct': {
        'callback': profile_distinct,
    },
    'map_reduce': {
        'callback': profile_map_reduce,
        'store_backtrace': True,
    },
    'inline_map_reduce': {
        'callback': profile_map_reduce,
        'store_backtrace': True,
    },
    'find_and_modify': {
        'callback': profile_find_and_modify,
        'store_backtrace': True,
    },
}

# For pymongo.database.Database
DATABASE_METHOD_INST = {
    'create_collection': {
        'callback': profile_create_collection,
    },
    'command': {
        'callback': profile_command,
    },
    'drop_collection': {
        'callback': profile_drop_collection,
    },
}

# For pymongo.cursor.Cursor
CURSOR_METHOD_INST = {
    'sort': {},
    'count': {},
    'distinct': {},
}


def wrap_class(cls, class_name, class_method_inst):
    """ wrap class methods with instrumentation calls """
    if not cls:
        return
    for (method, method_log_args) in class_method_inst.items():
        fn = getattr(cls, method, None)
        if not fn:
            # Not all methods may be in all versions of pymongo...
            continue
        kvs = {'Class': '%s.%s' % (cls.__module__, cls.__name__),
               'Function': method,
               'Action': '%s.%s' % (class_name, method),
               }
        setattr(cls, method, util.log_method(PYMONGO_LAYER, entry_kvs=kvs, **method_log_args)(fn))


def wrap(module):
    """ wrap pymongo module, adding instrumentation to core classes """
    try:
        wrap_class(module.collection.Collection, 'collection', COLLECTION_METHOD_INST)
        wrap_class(module.database.Database, 'database', DATABASE_METHOD_INST)
        wrap_class(module.cursor.Cursor, 'cursor', CURSOR_METHOD_INST)
    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM error: %s" % str(e))


if util.ready():
    try:
        import pymongo
        from appoptics_apm.json_encoder import JSONEncoder
        from appoptics_apm.skeleton import skeleton
        from bson.son import SON

        wrap(pymongo)
    except ImportError as e:
        pass



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 2.0.12
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2,6,0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_oboe', [dirname(__file__)])
        except ImportError:
            import _oboe
            return _oboe
        if fp is not None:
            try:
                _mod = imp.load_module('_oboe', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _oboe = swig_import_helper()
    del swig_import_helper
else:
    import _oboe
del version_info
try:
    _swig_property = property
except NameError:
    pass # Python < 2.2 doesn't have 'property'.
def _swig_setattr_nondynamic(self,class_type,name,value,static=1):
    if (name == "thisown"): return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name,None)
    if method: return method(self,value)
    if (not static):
        self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)

def _swig_setattr(self,class_type,name,value):
    return _swig_setattr_nondynamic(self,class_type,name,value,0)

def _swig_getattr(self,class_type,name):
    if (name == "thisown"): return self.this.own()
    method = class_type.__swig_getmethods__.get(name,None)
    if method: return method(self)
    raise AttributeError(name)

def _swig_repr(self):
    try: strthis = "proxy of " + self.this.__repr__()
    except: strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object : pass
    _newclass = 0


OBOE_DEBUG_FATAL = _oboe.OBOE_DEBUG_FATAL
OBOE_DEBUG_ERROR = _oboe.OBOE_DEBUG_ERROR
OBOE_DEBUG_WARNING = _oboe.OBOE_DEBUG_WARNING
OBOE_DEBUG_INFO = _oboe.OBOE_DEBUG_INFO
OBOE_DEBUG_LOW = _oboe.OBOE_DEBUG_LOW
OBOE_DEBUG_MEDIUM = _oboe.OBOE_DEBUG_MEDIUM
OBOE_DEBUG_HIGH = _oboe.OBOE_DEBUG_HIGH
OBOE_MODULE_ALL = _oboe.OBOE_MODULE_ALL
OBOE_MODULE_UNDEF = _oboe.OBOE_MODULE_UNDEF
OBOE_MODULE_LIBOBOE = _oboe.OBOE_MODULE_LIBOBOE
OBOE_MODULE_SETTINGS = _oboe.OBOE_MODULE_SETTINGS
OBOE_MODULE_REPORTER_FILE = _oboe.OBOE_MODULE_REPORTER_FILE
OBOE_MODULE_REPORTER_UDP = _oboe.OBOE_MODULE_REPORTER_UDP
OBOE_MODULE_REPORTER_SSL = _oboe.OBOE_MODULE_REPORTER_SSL
OBOE_MODULE_APACHE = _oboe.OBOE_MODULE_APACHE
OBOE_MODULE_NGINX = _oboe.OBOE_MODULE_NGINX
OBOE_MODULE_PHP = _oboe.OBOE_MODULE_PHP
OBOE_MODULE_DOTNET = _oboe.OBOE_MODULE_DOTNET
class Metadata(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Metadata, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Metadata, name)
    __repr__ = _swig_repr
    def __init__(self, *args): 
        this = _oboe.new_Metadata(*args)
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_Metadata
    __del__ = lambda self : None;
    __swig_getmethods__["fromString"] = lambda x: _oboe.Metadata_fromString
    if _newclass:fromString = staticmethod(_oboe.Metadata_fromString)
    def createEvent(self): return _oboe.Metadata_createEvent(self)
    __swig_getmethods__["makeRandom"] = lambda x: _oboe.Metadata_makeRandom
    if _newclass:makeRandom = staticmethod(_oboe.Metadata_makeRandom)
    def copy(self): return _oboe.Metadata_copy(self)
    def isValid(self): return _oboe.Metadata_isValid(self)
    def isSampled(self): return _oboe.Metadata_isSampled(self)
    def toString(self): return _oboe.Metadata_toString(self)
Metadata_swigregister = _oboe.Metadata_swigregister
Metadata_swigregister(Metadata)

def Metadata_fromString(*args):
  return _oboe.Metadata_fromString(*args)
Metadata_fromString = _oboe.Metadata_fromString

def Metadata_makeRandom(sampled=True):
  return _oboe.Metadata_makeRandom(sampled)
Metadata_makeRandom = _oboe.Metadata_makeRandom

class Context(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Context, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Context, name)
    __repr__ = _swig_repr
    __swig_getmethods__["setTracingMode"] = lambda x: _oboe.Context_setTracingMode
    if _newclass:setTracingMode = staticmethod(_oboe.Context_setTracingMode)
    __swig_getmethods__["setDefaultSampleRate"] = lambda x: _oboe.Context_setDefaultSampleRate
    if _newclass:setDefaultSampleRate = staticmethod(_oboe.Context_setDefaultSampleRate)
    __swig_getmethods__["sampleRequest"] = lambda x: _oboe.Context_sampleRequest
    if _newclass:sampleRequest = staticmethod(_oboe.Context_sampleRequest)
    __swig_getmethods__["get"] = lambda x: _oboe.Context_get
    if _newclass:get = staticmethod(_oboe.Context_get)
    __swig_getmethods__["toString"] = lambda x: _oboe.Context_toString
    if _newclass:toString = staticmethod(_oboe.Context_toString)
    __swig_getmethods__["set"] = lambda x: _oboe.Context_set
    if _newclass:set = staticmethod(_oboe.Context_set)
    __swig_getmethods__["fromString"] = lambda x: _oboe.Context_fromString
    if _newclass:fromString = staticmethod(_oboe.Context_fromString)
    __swig_getmethods__["copy"] = lambda x: _oboe.Context_copy
    if _newclass:copy = staticmethod(_oboe.Context_copy)
    __swig_getmethods__["setSampledFlag"] = lambda x: _oboe.Context_setSampledFlag
    if _newclass:setSampledFlag = staticmethod(_oboe.Context_setSampledFlag)
    __swig_getmethods__["clear"] = lambda x: _oboe.Context_clear
    if _newclass:clear = staticmethod(_oboe.Context_clear)
    __swig_getmethods__["isValid"] = lambda x: _oboe.Context_isValid
    if _newclass:isValid = staticmethod(_oboe.Context_isValid)
    __swig_getmethods__["isSampled"] = lambda x: _oboe.Context_isSampled
    if _newclass:isSampled = staticmethod(_oboe.Context_isSampled)
    __swig_getmethods__["init"] = lambda x: _oboe.Context_init
    if _newclass:init = staticmethod(_oboe.Context_init)
    __swig_getmethods__["init_reporter"] = lambda x: _oboe.Context_init_reporter
    if _newclass:init_reporter = staticmethod(_oboe.Context_init_reporter)
    __swig_getmethods__["shutdown"] = lambda x: _oboe.Context_shutdown
    if _newclass:shutdown = staticmethod(_oboe.Context_shutdown)
    __swig_getmethods__["isReady"] = lambda x: _oboe.Context_isReady
    if _newclass:isReady = staticmethod(_oboe.Context_isReady)
    __swig_getmethods__["createEvent"] = lambda x: _oboe.Context_createEvent
    if _newclass:createEvent = staticmethod(_oboe.Context_createEvent)
    __swig_getmethods__["startTrace"] = lambda x: _oboe.Context_startTrace
    if _newclass:startTrace = staticmethod(_oboe.Context_startTrace)
    def __init__(self): 
        this = _oboe.new_Context()
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_Context
    __del__ = lambda self : None;
Context_swigregister = _oboe.Context_swigregister
Context_swigregister(Context)

def Context_setTracingMode(*args):
  return _oboe.Context_setTracingMode(*args)
Context_setTracingMode = _oboe.Context_setTracingMode

def Context_setDefaultSampleRate(*args):
  return _oboe.Context_setDefaultSampleRate(*args)
Context_setDefaultSampleRate = _oboe.Context_setDefaultSampleRate

def Context_sampleRequest(*args):
  return _oboe.Context_sampleRequest(*args)
Context_sampleRequest = _oboe.Context_sampleRequest

def Context_get():
  return _oboe.Context_get()
Context_get = _oboe.Context_get

def Context_toString():
  return _oboe.Context_toString()
Context_toString = _oboe.Context_toString

def Context_set(*args):
  return _oboe.Context_set(*args)
Context_set = _oboe.Context_set

def Context_fromString(*args):
  return _oboe.Context_fromString(*args)
Context_fromString = _oboe.Context_fromString

def Context_copy():
  return _oboe.Context_copy()
Context_copy = _oboe.Context_copy

def Context_setSampledFlag():
  return _oboe.Context_setSampledFlag()
Context_setSampledFlag = _oboe.Context_setSampledFlag

def Context_clear():
  return _oboe.Context_clear()
Context_clear = _oboe.Context_clear

def Context_isValid():
  return _oboe.Context_isValid()
Context_isValid = _oboe.Context_isValid

def Context_isSampled():
  return _oboe.Context_isSampled()
Context_isSampled = _oboe.Context_isSampled

def Context_init(*args):
  return _oboe.Context_init(*args)
Context_init = _oboe.Context_init

def Context_init_reporter(*args):
  return _oboe.Context_init_reporter(*args)
Context_init_reporter = _oboe.Context_init_reporter

def Context_shutdown():
  return _oboe.Context_shutdown()
Context_shutdown = _oboe.Context_shutdown

def Context_isReady(*args):
  return _oboe.Context_isReady(*args)
Context_isReady = _oboe.Context_isReady

def Context_createEvent():
  return _oboe.Context_createEvent()
Context_createEvent = _oboe.Context_createEvent

def Context_startTrace():
  return _oboe.Context_startTrace()
Context_startTrace = _oboe.Context_startTrace

class Event(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Event, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Event, name)
    def __init__(self, *args, **kwargs): raise AttributeError("No constructor defined")
    __repr__ = _swig_repr
    __swig_destroy__ = _oboe.delete_Event
    __del__ = lambda self : None;
    def addInfo(self, *args): return _oboe.Event_addInfo(self, *args)
    def addEdge(self, *args): return _oboe.Event_addEdge(self, *args)
    def addEdgeStr(self, *args): return _oboe.Event_addEdgeStr(self, *args)
    def getMetadata(self): return _oboe.Event_getMetadata(self)
    def metadataString(self): return _oboe.Event_metadataString(self)
    def send(self): return _oboe.Event_send(self)
    __swig_getmethods__["startTrace"] = lambda x: _oboe.Event_startTrace
    if _newclass:startTrace = staticmethod(_oboe.Event_startTrace)
Event_swigregister = _oboe.Event_swigregister
Event_swigregister(Event)

def Event_startTrace(*args):
  return _oboe.Event_startTrace(*args)
Event_startTrace = _oboe.Event_startTrace

class Span(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Span, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Span, name)
    __repr__ = _swig_repr
    __swig_getmethods__["createSpan"] = lambda x: _oboe.Span_createSpan
    if _newclass:createSpan = staticmethod(_oboe.Span_createSpan)
    __swig_getmethods__["createHttpSpan"] = lambda x: _oboe.Span_createHttpSpan
    if _newclass:createHttpSpan = staticmethod(_oboe.Span_createHttpSpan)
    def __init__(self): 
        this = _oboe.new_Span()
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_Span
    __del__ = lambda self : None;
Span_swigregister = _oboe.Span_swigregister
Span_swigregister(Span)

def Span_createSpan(*args):
  return _oboe.Span_createSpan(*args)
Span_createSpan = _oboe.Span_createSpan

def Span_createHttpSpan(*args):
  return _oboe.Span_createHttpSpan(*args)
Span_createHttpSpan = _oboe.Span_createHttpSpan

class MetricTags(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, MetricTags, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, MetricTags, name)
    __repr__ = _swig_repr
    def __init__(self, *args): 
        this = _oboe.new_MetricTags(*args)
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_MetricTags
    __del__ = lambda self : None;
    def add(self, *args): return _oboe.MetricTags_add(self, *args)
MetricTags_swigregister = _oboe.MetricTags_swigregister
MetricTags_swigregister(MetricTags)

class CustomMetrics(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, CustomMetrics, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, CustomMetrics, name)
    __repr__ = _swig_repr
    __swig_getmethods__["summary"] = lambda x: _oboe.CustomMetrics_summary
    if _newclass:summary = staticmethod(_oboe.CustomMetrics_summary)
    __swig_getmethods__["increment"] = lambda x: _oboe.CustomMetrics_increment
    if _newclass:increment = staticmethod(_oboe.CustomMetrics_increment)
    def __init__(self): 
        this = _oboe.new_CustomMetrics()
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_CustomMetrics
    __del__ = lambda self : None;
CustomMetrics_swigregister = _oboe.CustomMetrics_swigregister
CustomMetrics_swigregister(CustomMetrics)

def CustomMetrics_summary(*args):
  return _oboe.CustomMetrics_summary(*args)
CustomMetrics_summary = _oboe.CustomMetrics_summary

def CustomMetrics_increment(*args):
  return _oboe.CustomMetrics_increment(*args)
CustomMetrics_increment = _oboe.CustomMetrics_increment

class Reporter(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Reporter, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Reporter, name)
    __repr__ = _swig_repr
    def __init__(self, *args): 
        this = _oboe.new_Reporter(*args)
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_Reporter
    __del__ = lambda self : None;
    def sendReport(self, *args): return _oboe.Reporter_sendReport(self, *args)
    def sendStatus(self, *args): return _oboe.Reporter_sendStatus(self, *args)
Reporter_swigregister = _oboe.Reporter_swigregister
Reporter_swigregister(Reporter)

class DebugLogger(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DebugLogger, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DebugLogger, name)
    def __init__(self, *args, **kwargs): raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _oboe.delete_DebugLogger
    __del__ = lambda self : None;
    def log(self, *args): return _oboe.DebugLogger_log(self, *args)
DebugLogger_swigregister = _oboe.DebugLogger_swigregister
DebugLogger_swigregister(DebugLogger)


def oboe_debug_log_handler(*args):
  return _oboe.oboe_debug_log_handler(*args)
oboe_debug_log_handler = _oboe.oboe_debug_log_handler
class DebugLog(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DebugLog, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DebugLog, name)
    __repr__ = _swig_repr
    __swig_getmethods__["getLevelName"] = lambda x: _oboe.DebugLog_getLevelName
    if _newclass:getLevelName = staticmethod(_oboe.DebugLog_getLevelName)
    __swig_getmethods__["getModuleName"] = lambda x: _oboe.DebugLog_getModuleName
    if _newclass:getModuleName = staticmethod(_oboe.DebugLog_getModuleName)
    __swig_getmethods__["getLevel"] = lambda x: _oboe.DebugLog_getLevel
    if _newclass:getLevel = staticmethod(_oboe.DebugLog_getLevel)
    __swig_getmethods__["setLevel"] = lambda x: _oboe.DebugLog_setLevel
    if _newclass:setLevel = staticmethod(_oboe.DebugLog_setLevel)
    __swig_getmethods__["setOutputStream"] = lambda x: _oboe.DebugLog_setOutputStream
    if _newclass:setOutputStream = staticmethod(_oboe.DebugLog_setOutputStream)
    __swig_getmethods__["setOutputFile"] = lambda x: _oboe.DebugLog_setOutputFile
    if _newclass:setOutputFile = staticmethod(_oboe.DebugLog_setOutputFile)
    __swig_getmethods__["addDebugLogger"] = lambda x: _oboe.DebugLog_addDebugLogger
    if _newclass:addDebugLogger = staticmethod(_oboe.DebugLog_addDebugLogger)
    __swig_getmethods__["removeDebugLogger"] = lambda x: _oboe.DebugLog_removeDebugLogger
    if _newclass:removeDebugLogger = staticmethod(_oboe.DebugLog_removeDebugLogger)
    __swig_getmethods__["logMessage"] = lambda x: _oboe.DebugLog_logMessage
    if _newclass:logMessage = staticmethod(_oboe.DebugLog_logMessage)
    def __init__(self): 
        this = _oboe.new_DebugLog()
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_DebugLog
    __del__ = lambda self : None;
DebugLog_swigregister = _oboe.DebugLog_swigregister
DebugLog_swigregister(DebugLog)

def DebugLog_getLevelName(*args):
  return _oboe.DebugLog_getLevelName(*args)
DebugLog_getLevelName = _oboe.DebugLog_getLevelName

def DebugLog_getModuleName(*args):
  return _oboe.DebugLog_getModuleName(*args)
DebugLog_getModuleName = _oboe.DebugLog_getModuleName

def DebugLog_getLevel(*args):
  return _oboe.DebugLog_getLevel(*args)
DebugLog_getLevel = _oboe.DebugLog_getLevel

def DebugLog_setLevel(*args):
  return _oboe.DebugLog_setLevel(*args)
DebugLog_setLevel = _oboe.DebugLog_setLevel

def DebugLog_setOutputStream(*args):
  return _oboe.DebugLog_setOutputStream(*args)
DebugLog_setOutputStream = _oboe.DebugLog_setOutputStream

def DebugLog_setOutputFile(*args):
  return _oboe.DebugLog_setOutputFile(*args)
DebugLog_setOutputFile = _oboe.DebugLog_setOutputFile

def DebugLog_addDebugLogger(*args):
  return _oboe.DebugLog_addDebugLogger(*args)
DebugLog_addDebugLogger = _oboe.DebugLog_addDebugLogger

def DebugLog_removeDebugLogger(*args):
  return _oboe.DebugLog_removeDebugLogger(*args)
DebugLog_removeDebugLogger = _oboe.DebugLog_removeDebugLogger

def DebugLog_logMessage(*args):
  return _oboe.DebugLog_logMessage(*args)
DebugLog_logMessage = _oboe.DebugLog_logMessage

class Config(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Config, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Config, name)
    __repr__ = _swig_repr
    __swig_getmethods__["checkVersion"] = lambda x: _oboe.Config_checkVersion
    if _newclass:checkVersion = staticmethod(_oboe.Config_checkVersion)
    __swig_getmethods__["getVersion"] = lambda x: _oboe.Config_getVersion
    if _newclass:getVersion = staticmethod(_oboe.Config_getVersion)
    __swig_getmethods__["getRevision"] = lambda x: _oboe.Config_getRevision
    if _newclass:getRevision = staticmethod(_oboe.Config_getRevision)
    def __init__(self): 
        this = _oboe.new_Config()
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _oboe.delete_Config
    __del__ = lambda self : None;
Config_swigregister = _oboe.Config_swigregister
Config_swigregister(Config)

def Config_checkVersion(*args):
  return _oboe.Config_checkVersion(*args)
Config_checkVersion = _oboe.Config_checkVersion

def Config_getVersion():
  return _oboe.Config_getVersion()
Config_getVersion = _oboe.Config_getVersion

def Config_getRevision():
  return _oboe.Config_getRevision()
Config_getRevision = _oboe.Config_getRevision

# This file is compatible with both classic and new-style classes.





""" AppOptics APM instrumentation API for Python.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.

appoptics_apm_noop defines no-op/test mock classes for:
a) platforms we don't support building the c extension on
b) running unit test
"""
import pprint

# global set of registered event reporting listeners
listeners = []


class Metadata(object):
    def __init__(self, _=None):
        pass

    @staticmethod
    def fromString(_):
        return Metadata()

    def createEvent(self):
        return Event()

    @staticmethod
    def makeRandom(flag=True):
        return Metadata(_=True)

    def copy(self):
        return self

    def isValid(self):
        return True

    def isSampled(self):
        return True

    def toString(self):
        return ''


class Context(object):
    md = None

    @staticmethod
    def init(_, __):
        pass

    @staticmethod
    def setTracingMode(_):
        return False

    @staticmethod
    def setDefaultSampleRate(_):
        return False

    @staticmethod
    def sampleRequest(_, __):
        return True

    @classmethod
    def get(cls):
        return cls.md

    @classmethod
    def set(cls, md):
        cls.md = md

    @staticmethod
    def fromString(_):
        return Context()

    @staticmethod
    def copy():
        return Context()

    @classmethod
    def clear(cls):
        cls.md = None

    @classmethod
    def isValid(cls):
        return cls.md != None

    @classmethod
    def isSampled(cls):
        return cls.md != None

    @staticmethod
    def toString():
        return ''

    @staticmethod
    def createEvent():
        return Event()

    @staticmethod
    def startTrace(_=None):
        return Event()


class Event(object):
    def __init__(self, _=None, __=None):
        self.props = {}

    def addInfo(self, name, value):
        self.props[name] = value

    def addEdge(self, _):
        pass

    def addEdgeStr(self, _):
        pass

    def getMetadata(self):
        return Metadata()

    def metadataString(self):
        return ''

    def is_valid(self):
        return True

    @staticmethod
    def startTrace(_=None):
        return Event()

    def __repr__(self):
        return '<Event('+str(self.props)+')>'


class Reporter(object):
    """ Mock the Reporter; no-op for unsupported platforms, or unit test harness
        if in APPOPTICS_TEST mode. """
    def __init__(self, _, __=None):
        pass

    def sendReport(self, event, __=None):
        for listener in listeners:
            listener.send(event)

    def sendStatus(self, event, __=None):
        for listener in listeners:
            listener.send(event)


class UdpReporter(object):
    """ Mock UDP Reported; no-op for unsupported platforms, or unit test harness
        if in APPOPTICS_TEST mode. """
    def __init__(self, _, __=None):
        pass

    def sendReport(self, event, __=None):
        for listener in listeners:
            listener.send(event)

    def sendStatus(self, event, __=None):
        for listener in listeners:
            listener.send(event)


class SslReporter(object):
    """ Mock UDP Reported; no-op for unsupported platforms, or unit test harness
        if in APPOPTICS_TEST mode. """
    def __init__(self, _, __=None):
        pass

    def sendReport(self, event, __=None):
        for listener in listeners:
            listener.send(event)

    def sendStatus(self, event, __=None):
        for listener in listeners:
            listener.send(event)

class Span(object):
    @staticmethod
    def createHttpSpan(*args):
        for listener in listeners:
            listener.send({'SPAN_REPORT': args})

    @staticmethod
    def createSpan(*args):
        for listener in listeners:
            listener.send({'SPAN_REPORT': args})

class MetricTags(object):
    def add(*args, **kwargs):
        pass


class CustomMetrics(object):
    @staticmethod
    def summary(*args, **kwargs):
        pass

    @staticmethod
    def increment(*args, **kwargs):
        pass


class oboe_metric_tag_t(object):
    def __init__(self, k, v):
        self.key = k
        self.value = v


class AppOpticsApmListener(object):
    """ Simple test harness for intercepting event reports. """
    def __init__(self):
        self.events = []
        self.spans = []
        self.listeners = listeners
        listeners.append(self)

    def send(self, event):
        if type(event) is Event: 
            self.events.append(event)
        else:
            self.spans.append(event)

    def get_events(self, *filters):
        """ Returns all events matching the filters passed """
        events = self.events
        for _filter in filters:
            events = [ev for ev in events if _filter(ev)]
        return events

    def get_spans(self, *filters):
        """ Returns all spans matching the filters passed """
        spans = self.spans
        for _filter in filters:
            spans = [sp for sp in spans if _filter(sp)]
        return spans

    def str_events(self, *filters):
        return pprint.pformat(self.get_events(*filters))

    def str_spans(self, *filters):
        return pprint.pformat(self.get_spans(*filters))

    def pop_events(self, *filters):
        """ Returns all events matching the filters passed,
        and also removes those events from the Trace so that
        they will not be returned by future calls to
        pop_events or events. """
        matched = self.get_events(*filters)
        for match in matched:
            self.events.remove(match)
        return matched

    def pop_spans(self, *filters):
        """ Returns all spans matching the filters passed,
        and also removes those spans from the Trace so that
        they will not be returned by future calls to
        pop_spans or spans. """
        matched = self.get_spans(*filters)
        for match in matched:
            self.spans.remove(match)
        return matched

    def __del__(self):
        self.listeners.remove(self)


class DebugLog(object):
    @staticmethod
    def getLevelName(*args, **kwargs):
        pass

    @staticmethod
    def getModuleName(*args, **kwargs):
        pass

    @staticmethod
    def getLevel(*args, **kwargs):
        pass

    @staticmethod
    def setLevel(*args, **kwargs):
        pass

    @staticmethod
    def setOutputStream(*args, **kwargs):
        pass

    @staticmethod
    def setOutputFile(*args, **kwargs):
        pass

    @staticmethod
    def addDebugLogger(*args, **kwargs):
        pass

    @staticmethod
    def removeDebugLogger(*args, **kwargs):
        pass

    @staticmethod
    def logMessage(*args, **kwargs):
        pass



""" AppOptics APM instrumentation API for Python.
When imported, the appoptics_apm_init() function will be executed to do some initialization tasks.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""

import logging
import inspect
import os
import time
import sys
import types
import traceback
from collections import defaultdict
import threading

import six
from decorator import decorator

__version__ = '3.4.1'

# "invalid name ... for type constant"
# pylint interprets all module-level variables as being 'constants'.
# pylint: disable=C0103

# Agent process start time, which is supposed to be a constant. -- don't modify it.
AGENT_START_TIME = time.time() * 1e6

# Constants from liboboe
OBOE_TRACE_NEVER = 0
OBOE_TRACE_ALWAYS = 1
OBOE_TRACE_THROUGH = 2

# oboe modules
OBOE_MODULE_ALL = -1

# Debug level from liboboe: oboe_debug.h
OBOE_DEBUG_FATAL = 0
OBOE_DEBUG_ERROR = 1
OBOE_DEBUG_WARNING = 2
OBOE_DEBUG_INFO = 3
OBOE_DEBUG_LOW = 4
OBOE_DEBUG_MEDIUM = 5
OBOE_DEBUG_HIGH = 6

# Map the oboe debug level to python logging level
_logging_map = {
    OBOE_DEBUG_FATAL: logging.CRITICAL,
    OBOE_DEBUG_ERROR: logging.ERROR,
    OBOE_DEBUG_WARNING: logging.WARNING,
    OBOE_DEBUG_INFO: logging.INFO,
    OBOE_DEBUG_LOW: logging.DEBUG,
    OBOE_DEBUG_MEDIUM: logging.DEBUG,
    OBOE_DEBUG_HIGH: logging.DEBUG
    }

# Sample sources
OBOE_SAMPLE_RATE_SOURCE_FILE = 1
OBOE_SAMPLE_RATE_SOURCE_DEFAULT = 2
OBOE_SAMPLE_RATE_SOURCE_OBOE = 3
OBOE_SAMPLE_RATE_SOURCE_LAST_OBOE = 4
OBOE_SAMPLE_RATE_SOURCE_DEFAULT_MISCONFIGURED = 5
OBOE_SAMPLE_RATE_SOURCE_OBOE_DEFAULT = 6
OBOE_SAMPLE_RATE_SOURCE_CUSTOM = 7

# Masks for bitwise ops
ZERO_MASK = 0b00000000000000000000000000
SAMPLE_RATE_MASK = 0b00111111111111111111111111
SAMPLE_SOURCE_MASK = 0b11000000000000000000000000
ZERO_SAMPLE_RATE_MASK = 0b11000000000000000000000000
ZERO_SAMPLE_SOURCE_MASK = 0b00111111111111111111111111

def _get_logger():
    """Define the logger.
    """
    if not (hasattr(logging, 'NullHandler') and callable(logging.NullHandler)):
        class NullHandler(logging.Handler):
            def emit(self, record):
                pass
        logging.NullHandler = NullHandler

    _logger = logging.getLogger(__name__)
    if 'APPOPTICS_APM_PYTHON_DISABLE_LOGGER' in os.environ or 'APPOPTICS_DISABLE_DEFAULT_LOGGER' in os.environ:
        _logger.addHandler(logging.NullHandler())
    else:
        sh = logging.StreamHandler() # get stream handler to custom logging prefix
        f = logging.Formatter('%(asctime)s [ %(name)s %(levelname)-8s] %(message)s')
        sh.setFormatter(f)
        _logger.addHandler(sh)  # use sys.stderr; see oboeware #63
    return _logger


logger = _get_logger()

reporter_instance = None

# test harness?  if not, check if we will run in no-op on this platform
TEST_MODE = 'APPOPTICS_TEST' in os.environ

if TEST_MODE:
    from appoptics_apm.appoptics_apm_test import (
        Context as SwigContext,
        Event as SwigEvent,
        Reporter,
        Metadata,
        DebugLog,
        Span as SwigSpan,
        CustomMetrics,
        MetricTags,
        )

    logger.error("AppOptics APM Oboe running in APPOPTICS_TEST mode; will not emit reports")
else:
    try:
        from appoptics_apm.swig.oboe import (
            Context as SwigContext,
            Event as SwigEvent,
            Reporter,
            Metadata,
            DebugLog,
            Span as SwigSpan,
            CustomMetrics,
            MetricTags,
            )
    except ImportError as e:
        from appoptics_apm.appoptics_apm_noop import (
            Context as SwigContext,
            Event as SwigEvent,
            Reporter,
            Metadata,
            DebugLog,
            Span as SwigSpan,
            CustomMetrics,
            MetricTags,
            )

        logger.warning("Module native extension lib not found. "
                       "Tracing is disabled and will go into no-op mode. "
                       "Contact support@appoptics.com if this is unexpected. "
                       "ImportError was: {e}".format(e=e))


def custom_metrics_summary(name, value, count=1, host_tag=False, service_name=None, tags=None, tags_count=0):
    """
    Creates a new or adds to an existing Summary Metric.

    :param name:str the name of the metrics, a part of the "metric key"
    :param value:float a value to be recorded associated with this "metric key"
    :param count:int (optional) count of metrics being reported, default is 1
    :param host_tag:boolean (optional) default is False, whether host information should be included
    :param service_name:str (optional) not yet supported, default is None
    :param tags: (optional) appoptics_apm.MetricTags which holds key/value pairs that describe the metric and a part of the "metric key", see example below
    :param tags_count: (optional) if tags is given, this must be set to the count of items in tags
    :return: None

    :example::
        # create a MetricTags object for two tags
        tags_count = 2
        tags = appoptics_apm.MetricTags(tags_count)
        # add tags into MetricTags by specifying the index and tag key and value
        tags.add(0, "Peter", "42")
        tags.add(1, "Paul", "45")
        # submit the metric
        appoptics_apm.custom_metrics_summary("my-summary-metric", 13.04, 1, False, None, tags, tags_count)
    """
    # convert None string to empty value, otherwise will cause oboe segfault
    if tags is None:
        tags = MetricTags(0); tags_count = 0
    CustomMetrics.summary(name, value, count, host_tag, None, tags, tags_count)


def custom_metrics_increment(name, count, host_tag=False, service_name=None, tags=None, tags_count=0):
    """
    Creates a new or adds to an existing Increment Metric.

    :param name:str the name of the metrics, a part of the "metric key"
    :param count:int increment value, typically 1
    :param host_tag:boolean (optional) default is False, whether host information should be included
    :param service_name:str (optional) not yet supported, default is None
    :param tags: (optional) appoptics_apm.MetricTags which holds key/value pairs that describe the metric and a part of the "metric key", see example below
    :param tags_count: (optional) if tags is given, this must be set to the count of items in tags
    :return: None

    :example::
        # create a MetricTags object for two tags
        tags_count = 2
        tags = appoptics_apm.MetricTags(tags_count)
        # add tags into MetricTags by specifying the index and tag key and value
        tags.add(0, "Peter", "42")
        tags.add(1, "Paul", "45")
        # submit the metric
        appoptics_apm.custom_metrics_increment("my-counter-metric", 1, False, None, tags, tags_count)
    """
    if tags is None:
        tags = MetricTags(0); tags_count = 0
    CustomMetrics.increment(name, count, host_tag, None, tags, tags_count)


def sys_is_traceable():
    """Check if the required system environment variables are all set."""
    # Python 2.6 has no set literals.
    required_envs = set([
        'APPOPTICS_SERVICE_KEY',
        ])
    return set(os.environ) & required_envs == required_envs


_ready = False


def ready():
    """This function is called by functions outside of this module.
    They should not check the value of _ready directly."""
    global _ready
    return _ready


def appoptics_apm_init():
    """Initialize the instrumentation context:
    1. Load the environment variables.
    2. Create a reporter, either ssl or udp
    3. Get sample configurations from collectors: sample_rate and sample_source"""
    global _ready

    if sys_is_traceable():
        # APPOPTICS_COLLECTOR is fetched by underlying library
        # Unfortunately, this function does not return anything for now.
        ret = SwigContext.init(os.environ.get('APPOPTICS_SERVICE_KEY'), os.environ.get('APPOPTICS_HOSTNAME_ALIAS', ''))
        if ret in (0, None):
            _ready = True
        else:
            logger.error(
                'Tracing enabled but initialization failed with err code: {ret}.'.format(ret=ret)
                )

    elif TEST_MODE:
        _ready = True
    else:
        logger.error(
            "APPOPTICS_SERVICE_KEY must be specified. Tracing disabled."
            )

"""The AppOptics APM initializer will be executed when this package is imported by user application"""
appoptics_apm_init()


class AppOpticsApmException(Exception):
    """ AppOptics APM Exception Class """
    pass


class AppOpticsApmConfig(object):
    """ AppOptics APM Configuration Class
    The precedence: Environment Variables > local configurations > default values
    For now liboboe will read the environment variables by itself so we don't need
    to refresh the settings to liboboe space, this applies to collector ip/port and
    service_key, as well as some other configurations."""
    complex_options = ['inst', 'transaction']
    additional_opts = ['sample_rate', 'tracing_mode',]
    delimiter = '.'

    def __init__(self, **kwargs):
        self._config = dict()

        # Update the config with default values
        self._config['collector_mode'] = 'ssl'  # ssl, udp
        self._config['enable_sanitize_sql'] = True  # Set to true to strip query literals
        self._config['reporter_host'] = '127.0.0.1'  # Report to localhost by default
        self._config['reporter_port'] = 4444  # hardcoded to 4444
        self._config['warn_deprecated'] = True
        self._config['inst_enabled'] = defaultdict(lambda: True)
        self._config['log_level'] = OBOE_DEBUG_ERROR

        self._config['transaction'] = defaultdict(lambda: True)
        self._config['transaction']['prepend_domain_name'] = False

        # Initialize dictionaries for per instrumentation configuration
        self._config['inst'] = defaultdict(lambda: True)

        self._config['inst']['django_orm'] = defaultdict(lambda: True)
        self._config['inst']['django_orm']['collect_backtraces'] = True

        self._config['inst']['httplib'] = defaultdict(lambda: True)
        self._config['inst']['httplib']['collect_backtraces'] = True

        self._config['inst']['memcache'] = defaultdict(lambda: True)
        self._config['inst']['memcache']['collect_backtraces'] = False

        self._config['inst']['pymongo'] = defaultdict(lambda: True)
        self._config['inst']['pymongo']['collect_backtraces'] = True

        self._config['inst']['redis'] = defaultdict(lambda: True)
        self._config['inst']['redis']['collect_backtraces'] = False

        self._config['inst']['sqlalchemy'] = defaultdict(lambda: True)
        self._config['inst']['sqlalchemy']['collect_backtraces'] = True

        cnf_file = os.environ.get('APPOPTICS_APM_CONFIG_PYTHON', os.environ.get('APPOPTICS_PYCONF', None))
        if cnf_file:
            self.update_with_cnf_file(cnf_file)

        self._config.update(kwargs)
        self.update_with_env_var()
        logger.debug('Intrumention config : {config}'.format(config = self._config))

    def update_with_env_var(self):
        """Update the settings with environment variables."""
        self._config['collector_mode'] = os.environ.get('APPOPTICS_REPORTER', self._config['collector_mode'])
        # We don't need to refresh the log_level for liboboe as it will read the same env variable
        # by itself. See oboe_debug_log_init()@ oboe_debug.c
        self._config['log_level'] = int(os.environ.get('APPOPTICS_DEBUG_LEVEL', self._config['log_level']))
        self._config['transaction']['prepend_domain_name'] = int(os.environ.get('APPOPTICS_APM_PREPEND_DOMAIN_NAME', self._config['transaction']['prepend_domain_name']))

    def update_with_cnf_file(self, cnf_path):
        """Update the settings with the config file, if any."""
        def _convert(value):
            """Convert vals to expected types."""
            value = value.lower()
            if value in ('true', 'false'):
                return True if value == 'true' else False
            try:
                value = float(value)
                return value
            except ValueError:
                return value

        cnf = six.moves.configparser.ConfigParser()
        try:
            if not cnf.read(cnf_path):
                logger.warning('Failed to open the config file: {}.'.format(cnf_path))
                return
        except six.moves.configparser.Error as e:
            logger.warning('Failed to read or parse config file: {e}'.format(e=e))

        allowed_opts = list(self._config.keys()) + self.additional_opts
        try:
            for section in cnf.sections():
                if section == 'main':
                    for key, val in cnf.items(section):
                        if key not in allowed_opts or key in self.complex_options:
                            continue
                        val = _convert(val)
                        # There is no 'sample_rate' and 'tracing_mode' in AppOpticsApmConfig so we cannot enforce
                        # type(self._config[key]) here.
                        self._config[key] = val
                elif section in self.complex_options:
                    for key, val in cnf.items(section):
                        if self.delimiter in key:
                            key, sub_key = key.split(self.delimiter, 1)
                        else:
                            sub_key = None
                        if key not in self._config[section]:
                            continue
                        val = _convert(val)
                        if sub_key and sub_key in self._config[section][key]:
                            self._config[section][key][sub_key] = type(self._config[section][key][sub_key])(val)
                        elif not sub_key:
                            self._config[section][key] = type(self._config[section][key])(val)
                        else:
                            logger.debug('Invalid key: [{s}] {k}.{sk}'.format(s=section, k=key, sk=sub_key))
                else:
                    logger.debug('Invalid section: {s} found in {f}'.format(s=section, f=cnf_path))
        except Exception as e:
            logger.warning('AppOptics APM config error: {e}'.format(e=e))

    def __setitem__(self, key, value):
        """Refresh the configurations in liboboe global struct while user changes settings.
        """
        if key == 'tracing_mode':
            if value == 'never':
                SwigContext.setTracingMode(0)
            elif value == 'always':
                SwigContext.setTracingMode(1)
            else:
                logger.warning('Unsupported trace mode: {mode}'.format(mode=value))
                return

            self._config[key] = value

        elif key == 'sample_rate':
            if not (isinstance(value, (int, float)) and (0 <= value <= 1)):
                logger.warning(
                    'Invalid type or range: {var} in {v_type}'.format(var=value, v_type=type(value))
                )
                return

            self._config[key] = value
            SwigContext.setDefaultSampleRate(int(value * 1e6))
            self._config['sample_source'] = OBOE_SAMPLE_RATE_SOURCE_FILE

        elif key in ('enable_sanitize_sql', 'warn_deprecated', 'sample_source'):
            self._config[key] = value
        else:
            logger.warning('Unsupported AppOptics APM config key: {key}'.format(key=key))

    def __getitem__(self, key):
        return self._config[key]

    def __delitem__(self, key):
        del self._config[key]

    def get(self, key, default=None):
        """ Get the value of key"""
        return self._config.get(key, default)


config = AppOpticsApmConfig()


###############################################################################
# Low-level Public API
###############################################################################

def _str_backtrace(backtrace=None):
    """ Return a string representation of an existing or new backtrace """
    if backtrace:
        return "".join(traceback.format_tb(backtrace))
    else:
        return "".join(traceback.format_stack()[:-1])


def _collect_backtraces(module_name):
    """ Return the collect backtraces config value for module """
    return config['inst'][module_name]['collect_backtraces']


class Context(object):
    """ A wrapper around the swig Metadata """

    #class scope dict to manage active_instance, transaction_name, and start_time
    #initialize it here (early) in case it is referenced before starting a trace
    transaction_dict = threading.local()
    def __init__(self, md):
        if isinstance(md, six.string_types):
            self._md = Metadata.fromString(md)
        else:
            self._md = md
        self.layer= None

    @property
    def md(self):
        return self._md

    # For interacting with SRv1

    @classmethod
    def set_tracing_mode(cls, mode):
        """ Updates liboboe with the configured tracing_mode """
        SwigContext.setTracingMode(mode)

    @classmethod
    def set_default_sample_rate(cls, rate):
        """ Updates liboboe with the configured sample_rate """
        SwigContext.setDefaultSampleRate(rate)

    # For interacting with the thread-local Context

    @classmethod
    def get_default(cls):
        """Returns the Context currently stored as the thread-local default."""
        return cls(SwigContext)

    def set_as_default(self):
        """Sets this object as the thread-local default Context.
        For now the liboboe does not check the validity of the context, it stores a context even
        the option byte is set to not tracing."""
        # We use is_valid() here as we should allow a context with flags=not tracing to be stored
        # in the context thread local storage.
        if self.is_valid():
            SwigContext.set(self._md)

    @classmethod
    def clear_default(cls):
        """Removes the current thread-local Context."""
        SwigContext.clear()

    @classmethod
    def xtr_is_v2(cls, xtr):
        """Check if the X-Trace ID string is V2"""
        return xtr and xtr[:2] == '2B'  # TODO: change to bitwise operations if more flags are added in future

    # For starting/stopping traces
    @classmethod
    def start_trace(cls, layer, xtr=None, force=False):
        """Returns a Context and a start event.
        Takes sampling into account -- may return an (invalid Context, event) pair.
        """
        Context.transaction_dict = threading.local()
        if sample_request(layer, xtr) or force:
            if cls.xtr_is_v2(xtr):
                xtr = '{ver_task_op}{flag}'.format(ver_task_op=xtr[:-2], flag='01')
                md = Metadata.fromString(xtr)
                evt = md.createEvent()
            else:
                md = Metadata.makeRandom(True)
                evt = SwigEvent.startTrace(md)
        else:
            evt = None
            if cls.xtr_is_v2(xtr):
                xtr = '{ver_task_op}{flag}'.format(ver_task_op=xtr[:-2], flag='00')
                md = Metadata.fromString(xtr)
            else:
                md = Metadata.makeRandom(False)

        if evt:
            event = Event(evt, 'entry', layer)
            sample_rate = getattr(Context.transaction_dict, 'sample_rate', config.get('sample_rate'))
            sample_source = getattr(Context.transaction_dict, 'sample_source', config.get('sample_source'))
            if sample_rate:
                event.add_info(
                    'SampleSource',
                    sample_source
                    )
                event.add_info('SampleRate', int(sample_rate * 1e6))
            event.add_info('Language', 'Python')  # in case we have an unclear layer name
        else:
            event = NullEvent()
        ctx = cls(md)
        ctx.layer= layer
        ctx.set_as_default()
        setattr(Context.transaction_dict, 'start_time', time.time() * 1e6)
        setattr(Context.transaction_dict, 'layer', layer)
        logger.debug('start_trace event : {ev}'.format(ev = event))
        return ctx, event

    def set_transaction_name(self, trans_name = ""):
        """ Set transaction name to this trace instance

           Keyword and  arguments:
               trans_name -- transaction name to tag this trace instance
        """

        if not trans_name:
            logger.warning('Invalid transaction name:{n}'.format( n = str(trans_name)))
            return False

        if type(trans_name) is not str:
            trans_name = str(trans_name)
        if self.is_valid():
            setattr(Context.transaction_dict, 'transaction_name', trans_name)
            return True
        else:
            logger.debug("Transactin name ignored as current execution flow is not monitored by agent")
        return False

    def get_transaction_name(self):
        """ Get transaction name of this trace instance"""

        return getattr(Context.transaction_dict, 'transaction_name', None)

    def end_trace(self, event):  # Reports the last event in a trace
        """Ends this trace, rendering this Context invalid.
           submit inbound metrics and get final transaction name
           update event info with transaction name

           Keyword arguments:
           self -- this context instance
           event -- last trace event
        """
        end_time = time.time() * 1e6
        domain = getattr(Context.transaction_dict, 'domain', None)
        if not config['transaction']['prepend_domain_name']:
            domain = None
        if domain is not None and len(domain.split(":")) > 1:
            port  = domain.split(":")[1]
            if port == "80" or port == "443":
                   domain = domain.split(":")[0]
        trans_name = getattr(Context.transaction_dict, 'transaction_name', None)
        http_span = getattr(Context.transaction_dict, 'http', None)
        request_method = getattr(Context.transaction_dict, 'request_method', None)
        status_code = getattr(Context.transaction_dict, 'status_code', None)

        start_time = getattr(Context.transaction_dict, 'start_time', None)
        if start_time is None:
            start_time = end_time
            logger.warning('trace start_time not recorded')
        else:
            start_time = int(start_time)
        url_tran = None if trans_name else getattr(Context.transaction_dict, 'url_tran', None)
        span_time = int(end_time - start_time)

        try:
            if http_span:
                if request_method is None or status_code is None:
                    logger.warning('http_span request info error request_method : {r}, status_code: {s}'.format(r = request_method, s = status_code))
                if trans_name is None and url_tran is None:
                    logger.warning('http_span error: transaction_name is None, use set_transaction_name("transaction_name") to set')
                if status_code is None:
                    status_code = 0
                has_error = bool(499 < status_code)  and bool(status_code < 600)
                trans_name = SwigSpan.createHttpSpan(trans_name, url_tran, domain, span_time, status_code, request_method, has_error)
            else:
                trans_name = SwigSpan.createSpan(trans_name, domain, span_time)
            logger.debug('end_trace create span: http: {http}, transaction_name :{xn}'.format(http = http_span, xn = trans_name))
        except Exception as e:
            logger.error('appoptics_apm::end_trace create SPAN fail: {0}'.format(e))

        logger.debug('end_trace event : {ev}'.format(ev = event))
        # True event is sampled, report it. Otherwise it's NullEvent or None, ignore it.
        if type(event) is Event:
            event.add_info('TransactionName', trans_name)
            if http_span:
                event.add_info('Method', request_method)
                event.add_info('HTTP-Host', getattr(Context.transaction_dict, 'domain', None))
                event.add_info('URL', getattr(Context.transaction_dict, 'url_tran', None))
                event.add_info('Status', status_code)
            self.report(event)
        self._md = None
        Context.clear_default()

    def create_event(self, label, layer):
        """Returns an Event associated with this Context."""
        if self.is_sampled():
            return Event(self._md.createEvent(), label, layer)
        else:
            return NullEvent()

    def report(self, event):
        """Report this Event.
        -------------------------------------------------------------------------------
        | Please note that we don't need to update the current op_id to the context   |
        | as the liboboe helps to do that in functions: oboe_send_event() -->         |
        | oboe_ids_set_op_id(). There is a risk for now that if the reporter fails to |
        | send the event out, the chain will be broken.                               |
        -------------------------------------------------------------------------------
        """
        if self.is_sampled() and event.is_valid():
            if self._md == SwigContext:
                _reporter().sendReport(event.evt)
            else:
                _reporter().sendReport(event.evt, self._md)
            logger.debug('report event: {e}'.format(e = event))

    def report_status(self, event):
        """Report with postStatus() Thrift API. It's mainly for the __Init message
        A code refactoring may be needed here for the boilerplate code"""
        if self.is_sampled() and event.is_valid():
            if self._md == SwigContext:
                _reporter().sendStatus(event.evt)
            else:
                _reporter().sendStatus(event.evt, self._md)

    def is_valid(self):
        """Returns whether this Context is valid.

        Call this before doing expensive introspection. If this returns False,
        then the context is not valid for moving forward to next step.
        """
        return self._md and self._md.isValid()

    def is_sampled(self):
        """Returns whether this Context is sampled.

        Call this before doing expensive introspection. If this returns False,
        then any event created by this Context will not actually return
        information to AppOptics APM.
        """
        return self.is_valid() and self._md.isSampled()

    def copy(self):
        """Make a clone of this Context."""
        return self.__class__(self._md.toString())

    def __str__(self):
        if self._md:
            return self._md.toString()
        else:
            return ''


class Event(object):
    """An Event is a key/value bag that will be reported to the Tracelyzer."""
    def __init__(self, raw_evt, label, layer):
        self._evt = raw_evt
        self._evt.addInfo('Label', label)
        self._evt.addInfo('Layer', layer)

    @property
    def evt(self):
        return self._evt

    def add_edge(self, ctx):
        """Connect an additional Context to this Event.

        All Events are created with an edge pointing to the previous Event. This
        creates an additional edge. This pattern is useful for entry/exit pairs
        in a layer.
        """
        if ctx.md == SwigContext:
            self._evt.addEdge(ctx.md.get())
        else:
            self._evt.addEdge(ctx.md)

    def add_edge_str(self, xtr):
        """Adds an edge to this Event, based on a str(Context).

        Useful for continuing a trace, e.g., from an X-Trace header in a service
        call.
        """
        self._evt.addEdgeStr(xtr)

    def add_info(self, key, value):
        """Add a key/value pair to this event."""
        self._evt.addInfo(key, value)

    def add_backtrace(self, backtrace=None):
        """Add a backtrace to this event.

        If backtrace is None, grab the backtrace from the current stack trace.
        """
        self.add_info('Backtrace', _str_backtrace(backtrace))

    @staticmethod
    def is_valid():
        """Returns whether this event will be reported to the Tracelyzer."""
        return True

    def id(self):
        """Returns a string version of this Event.

        Useful for attaching to output service calls (e.g., an X-Trace request
        header).
        """
        return self._evt.metadataString()


class NullEvent(object):
    """Subclass of event that will not be reported to the Tracelyzer.

    All methods here are no-ops. Checking for this class can be done
    (indirectly) by calling is_valid() on an object.
    """

    def __init__(self):
        pass

    def add_edge(self, event):
        pass

    def add_edge_str(self, op_id):
        pass

    def add_info(self, key, value):
        pass

    def add_backtrace(self, backtrace=None):
        pass

    @staticmethod
    def is_valid():
        return False

    @staticmethod
    def id():
        return ''


###############################################################################
# High-level Public API
###############################################################################

try:
    if six.PY2:
        from cStringIO import StringIO
    else:
        from io import StringIO

    import cProfile, pstats
    found_cprofile = True
except ImportError:
    found_cprofile = False


def _get_profile_info(p):
    """Returns a sorted set of stats from a cProfile instance."""
    sio = StringIO()
    s = pstats.Stats(p, stream=sio)
    s.sort_stats('time')
    s.print_stats(15)
    stats = sio.getvalue()
    sio.close()
    return stats

def _update_event(evt, keys=None, store_backtrace=True, backtrace=None, edge_str=None):
    """Add the backtrace and edge to the event
    """
    keys = keys or {}

    for k, v in keys.items():
        evt.add_info(k, v)

    if store_backtrace:
        evt.add_backtrace(backtrace)

    if edge_str:
        evt.add_edge_str(edge_str)
    return evt

def _log_event(evt, keys=None, store_backtrace=True, backtrace=None, edge_str=None):
    """Add the backtrace and edge, then send it out to the reporter.
    """
    evt = _update_event(evt, keys, store_backtrace, backtrace, edge_str)
    ctx = Context.get_default()
    ctx.report(evt)


def log(label, layer, keys=None, store_backtrace=True, backtrace=None, edge_str=None):
    """Report a single tracing event.

    :label: 'entry', 'exit', 'info', or 'error'
    :layer: The layer name
    :keys: A optional dictionary of key-value pairs to report.
    :store_backtrace: Whether to report a backtrace. Default: True
    :backtrace: The backtrace to report. Default: this call.
    """
    ctx = Context.get_default()
    if not ctx.is_sampled():
        return
    evt = ctx.create_event(label, layer)
    _log_event(
        evt, keys=keys, store_backtrace=store_backtrace,
        backtrace=backtrace, edge_str=edge_str
    )


def start_trace(layer, xtr=None, keys=None, store_backtrace=True, backtrace=None):
    """Start a new trace, or continue one from an external layer.

    :layer: The layer name of the root of the trace.
    :xtr: The X-Trace ID to continue this trace with.
    :keys: An optional dictionary of key-value pairs to report.
    :store_backtrace: Whether to report a backtrace. Default: True
    :backtrace: The backtrace to report. Default: this call.
    """
    keys = keys or {}
    forced_trace = 'Force' in keys

    ctx, evt = Context.start_trace(layer, xtr=xtr, force=forced_trace)

    if ctx.is_valid():  # Set it to the thread local storage even it's not sampled.
        ctx.set_as_default()

    if ctx.is_sampled():
        _log_event(evt, keys=keys, store_backtrace=store_backtrace, backtrace=backtrace)

    return ctx

def start_http_trace(layer, xtr=None, keys=None, store_backtrace=True, backtrace=None, transaction_name = None):
    """Start a new http trace, or continue one from an external layer.

    :layer: The layer name of the root of the trace.
    :xtr: The X-Trace ID to continue this trace with.
    :keys: An optional dictionary of key-value pairs to report.
    :store_backtrace: Whether to report a backtrace. Default: True
    :backtrace: The backtrace to report. Default: this call.
    """
    ctx = start_trace(layer, xtr, keys, store_backtrace, backtrace)
    if ctx.is_valid() and transaction_name is not None:
        ctx.set.transaction_name(transaction_name)
    return ctx



def end_trace(layer, keys=None, http = False):
    """End a trace, reporting a final event.

    This will end a trace locally. If the X-Trace ID returned here is reported
    externally, other processes can continue this trace.

    :layer: The layer name of the final layer.
    :keys: An optional dictionary of key-value pairs to report.
    :http: A flag indicates http span for True or non for False
    """
    setattr(Context.transaction_dict, 'http', http )
    if not http:
        setattr(Context.transaction_dict, 'domain', None )
        setattr(Context.transaction_dict, 'request_method', None)

    ctx = Context.get_default()
    evt = ctx.create_event('exit', layer)
    logger.debug('end_trace create event: {e}'.format(e = evt))
    if not ctx.is_sampled():
        ctx.end_trace(evt)
        return str(ctx)  # Always return the context id in string format.
    else:
        evt = _update_event(evt, keys=keys, store_backtrace=False)
        ctx_id = last_id()
        ctx.end_trace(evt)

    return ctx_id

def end_http_trace(layer_or_ctx, keys=None):
    """End a trace, reporting a final event.

    This will end a trace locally. If the X-Trace ID returned here is reported
    externally, other processes can continue this trace.

    :layer: The layer name of the final layer.
    :keys: An optional dictionary of key-value pairs to report.
    """
    setattr(Context.transaction_dict, 'http', True)

    if type(layer_or_ctx) is str:
        return end_trace(layer_or_ctx, keys, True)
    logger.error("AppOptics APM layer name not str type")
    return end_trace(Context.transaction_dict.get('layer', "unknow"), keys, True)

def set_request_info(host = None, status_code = None, method = None, full_path = None):
    """set or update http span info for current request

      this setting only take effect if currently there is an actve tracing
      instance

      Keywords and parameters:
          host -- request host name
          status_code -- response status code
          method -- request method
          full_path -- full uri
    """

    if status_code is not None:
        setattr(Context.transaction_dict, 'status_code', status_code)
    if method is not None:
        setattr(Context.transaction_dict, 'request_method', method)
    if host is not None:
        setattr(Context.transaction_dict, 'domain', host)
    if full_path is not None:
        setattr(Context.transaction_dict, 'url_tran', full_path)

def log_entry(layer, keys=None, store_backtrace=True, backtrace=None):
    """Report the first event of a new layer.

    :layer: The layer name.
    :keys: An optional dictionary of key-value pairs to report.
    :store_backtrace: Whether to report a backtrace. Default: True
    :backtrace: The backtrace to report. Default: this call.
    """
    ctx = Context.get_default()
    if not ctx.is_sampled():
        return
    evt = ctx.create_event('entry', layer)
    _log_event(evt, keys=keys, store_backtrace=store_backtrace, backtrace=backtrace)


def log_error(err_class, err_msg, store_backtrace=True, backtrace=None):
    """Report an error event.

    :err_class: The class of error to report, e.g., the name of the Exception.
    :err_msg: The specific error that occurred.
    :store_backtrace: Whether to report a backtrace. Default: True
    :backtrace: The backtrace to report. Default: this call.
    """
    ctx = Context.get_default()
    if not ctx.is_sampled():
        return
    evt = ctx.create_event('error', None)
    logger.debug('log_error create event: {e}'.format(e = evt))
    keys = {
        'ErrorClass': err_class,
        'ErrorMsg': err_msg
        }
    _log_event(evt, keys=keys, store_backtrace=store_backtrace, backtrace=backtrace)


def log_exception(msg=None, store_backtrace=True):
    """Report the last thrown exception as an error

    :msg: An optional message, to override err_msg. Defaults to str(Exception).
    :store_backtrace: Whether to store the Exception backtrace.
    """
    typ, val, tb = sys.exc_info()
    try:
        if typ is None:
            logger.debug(
                'log_exception should only be called from an exception context '
                '(e.g., except: block)'
            )
            return

        if msg is None:
            try:
                msg = str(val)
            except Exception:
                msg = repr(val)

        log_error(typ.__name__, msg,
                  store_backtrace=store_backtrace,
                  backtrace=tb if store_backtrace else None)
    finally:
        del tb  # delete reference to traceback object to allow garbage collection


def log_exit(layer, keys=None, store_backtrace=True, backtrace=None, edge_str=None):
    """Report the last event of the current layer.

    :layer: The layer name.
    :keys: An optional dictionary of key-value pairs to report.
    :store_backtrace: Whether to report a backtrace. Default: True
    :backtrace: The backtrace to report. Default: this call.
    """
    ctx = Context.get_default()
    if not ctx.is_sampled():
        return
    evt = ctx.create_event('exit', layer)
    logger.debug('log_exit create event: {e}'.format(e = evt))
    _log_event(
        evt, keys=keys, store_backtrace=store_backtrace, backtrace=backtrace, edge_str=edge_str
    )

def set_transaction_name(trans_name =""):
    """ Sets a transaction name to the current active trace, the transaction name will
    be reported along with the corresponding trace and metrics.
    This overrides the transaction name provided by out-of-the-box instrumentation.
    If multiple transaction names are set on the same trace, then the last one would be used.
    Take note that transaction name might be truncated with invalid characters replaced.

    Keyword arguments:
    trans_name -- customer defined transaction name
    return -- result of setting, True for success
    """

    ctx = Context.get_default()
    ret = ctx.set_transaction_name(trans_name)
    logger.debug('set_transaction_name:{n} returns  {r}'.format(n = trans_name, r = ret))
    return ret

def get_transaction_name():
    """ Returns the currently set custom transaction name, if any. """

    ctx = Context.get_default()
    return ctx.get_transaction_name()

def last_id():
    """Returns a string representation the last event reported."""
    return str(Context.get_default())


###############################################################################
# Python-specific functions
###############################################################################


def _function_signature(func):
    """Returns a string representation of the function signature of the given func."""
    name = func.__name__
    (args, varargs, keywords, defaults) = inspect.getargspec(func)
    argstrings = []
    if defaults:
        first = len(args) - len(defaults)
        argstrings = args[:first]
        for i in range(first, len(args)):
            d = defaults[i - first]
            if isinstance(d, six.string_types):
                d = "'" + d + "'"
            else:
                d = str(d)
            argstrings.append(args[i] + '=' + d)
    else:
        argstrings = args
    if varargs:
        argstrings.append('*' + varargs)
    if keywords:
        argstrings.append('**' + keywords)
    return name + '(' + ', '.join(argstrings) + ')'


def trace(layer='Python', xtr_hdr=None, kvs=None):
    """ Decorator to begin a new trace on a block of code.  Takes into account
    appoptics_apm.config['tracing_mode'] as well as appoptics_apm.config['sample_rate'], so may
    not always start a trace.

    :layer: layer name to report as
    :xtr_hdr: optional, incoming x-trace header if available
    :kvs: optional, dictionary of additional key/value pairs to report
    """

    def _trace_wrapper(func, *f_args, **f_kwargs):
        start_trace(layer, keys=kvs, xtr=xtr_hdr)
        try:
            res = func(*f_args, **f_kwargs)
        except Exception:
            # log exception and re-raise
            log_exception()
            raise
        finally:
            end_trace(layer)

        return res  # return output of func(*f_args, **f_kwargs)

    _trace_wrapper._appoptics_apm_wrapped = True  # mark our wrapper for protection below

    # instrumentation helper decorator, called to add wrapper at "decorate" time
    def decorate_with_trace(f):
        if getattr(f, '_appoptics_apm_wrapped', False):  # has this function already been wrapped?
            return f  # then pass through
        return decorator(_trace_wrapper, f)  # otherwise wrap function f with wrapper

    return decorate_with_trace


class profile_block(object):
    """A context manager for AppOptics APM profiling a block of code with AppOptics APM lib.

    Reports an error event between entry and exit if an exception is thrown,
    then reraises.

    :profile_name: the profile name to use when reporting.  this should be
        unique to the profiled method.
    :store_backtrace: whether to capture a backtrace or not (False)
    :profile: profile this function with cProfile and report the result
    """

    def __init__(self, profile_name, profile=False, store_backtrace=False):
        self.profile_name = profile_name
        self.use_cprofile = profile
        self.backtrace = store_backtrace
        self.p = None  # possible cProfile.Profile() instance

    def __enter__(self):
        ctx = Context.get_default()
        if not ctx.is_sampled():
            return

        # build entry event
        entry_kvs = {
            'Language': 'python',
            'ProfileName': self.profile_name,
            # XXX We can definitely figure out a way to make these
            # both available and fast.  For now, this is ok.
            'File': '',
            'LineNumber': 0,
            'Module': '',
            'FunctionName': '',
            'Signature': ''
            }
        log('profile_entry', None, keys=entry_kvs, store_backtrace=self.backtrace)

        # begin profiling
        if self.use_cprofile and found_cprofile:
            self.p = cProfile.Profile()
            self.p.enable(subcalls=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        ctx = Context.get_default()
        if not ctx.is_sampled():
            return

        # end profiling
        stats = None
        if self.use_cprofile and found_cprofile and self.p:
            stats = _get_profile_info(self.p)

        # exception?
        if exc_type:
            log_exception()

        # build exit event
        exit_kvs = {}
        if self.use_cprofile and stats:
            exit_kvs['ProfileStats'] = stats
        exit_kvs['Language'] = 'python'
        exit_kvs['ProfileName'] = self.profile_name

        log('profile_exit', None, keys=exit_kvs, store_backtrace=self.backtrace)


def profile_function(profile_name, store_args=False, store_return=False, store_backtrace=False,
                     profile=False, callback=None, entry_kvs=None):
    """Wrap a method for tracing and profiling with the AppOptics APM library.

    Reports an error event between entry and exit if an exception is thrown,
    then reraises.

    :profile_name: the profile name to use when reporting.  this should be
        unique to the profiled method.
    :store_return: report the return value of this function
    :store_args: report the arguments to this function
    :store_backtrace: whether to capture a backtrace or not (False)
    :profile: profile this function with cProfile and report the result
    :callback: if set, calls this function after the wrapped function returns,
        which examines the function, arguments, and return value, and may add
        more K/V pairs to the dictionary to be reported
    """

    def before(func, f_args, f_kwargs):
        # get filename, line number, etc, and cache in wrapped function to avoid overhead
        def cache(name, value_func):
            try:
                if not hasattr(func, name):
                    setattr(func, name, value_func())
            except Exception:
                setattr(func, name, None)

        cache('_appoptics_apm_file', lambda: inspect.getsourcefile(func))
        cache('_appoptics_apm_line_number', lambda: inspect.getsourcelines(func)[1])
        cache('_appoptics_apm_module', lambda: inspect.getmodule(func).__name__)
        cache('_appoptics_apm_signature', lambda: _function_signature(func))

        keys = {
            'Language': 'python',
            'ProfileName': profile_name,
            'File': getattr(func, '_appoptics_apm_file'),
            'LineNumber': getattr(func, '_appoptics_apm_line_number'),
            'Module': getattr(func, '_appoptics_apm_module'),
            'FunctionName': getattr(func, '__name__'),
            'Signature': getattr(func, '_appoptics_apm_signature')
        }
        return f_args, f_kwargs, keys

    def after(func, f_args, f_kwargs, res):

        kvs = {
            'Language': 'python',
            'ProfileName': profile_name
        }

        if callback:
            user_kvs = callback(func, f_args, f_kwargs, res)
            if user_kvs:
                kvs.update(user_kvs)

        return kvs

    # Do function passed in here expect to be bound (have im_func/im_class)?

    return log_method(
        None, store_return=store_return, store_args=store_args, store_backtrace=store_backtrace,
        before_callback=before, callback=after, profile=profile, entry_kvs=entry_kvs
    )


def log_method(layer, store_return=False, store_args=False, store_backtrace=False,
               before_callback=None, callback=None, profile=False, entry_kvs=None,
               send_entry_event=True, send_exit_event=True):
    """Wrap a method for tracing with the AppOptics APM library.

    As opposed to profile_function, this decorator gives the method its own layer

    Reports an error event between entry and exit if an exception is thrown,
    then reraises.

    :layer: the layer to use when reporting. If none, this layer will be a
        profile.
    :store_return: report the return value
    :store_args: report the arguments to this function
    :before_callback: if set, calls this function before the wrapped function is
        called. This function can change the args and kwargs, and can return K/V
        pairs to be reported in the entry event.
    :callback: if set, calls this function after the wrapped function returns,
        which examines the function, arguments, and return value, and may add
        more K/V pairs to the dictionary to be reported
    """
    if not entry_kvs:
        entry_kvs = {}

    # run-time event-reporting function, called at each invocation of func(f_args, f_kwargs)
    def _log_method_wrapper(func, *f_args, **f_kwargs):
        ctx = Context.get_default()
        #ctx should never be None, checking here in case
        if ctx is None or not ctx.is_sampled() or not ready():
            if ctx is None:
                logger.error("util._log_method_wrapper failed to get default context")
            return func(*f_args, **f_kwargs)  # pass through to func right away
        if store_args:
            entry_kvs.update({'args': f_args, 'kwargs': f_kwargs})
        if before_callback:
            before_res = before_callback(func, f_args, f_kwargs)
            if before_res:
                f_args, f_kwargs, extra_entry_kvs = before_res
                entry_kvs.update(extra_entry_kvs)
        if store_backtrace:
            entry_kvs['Backtrace'] = _str_backtrace()
        # is func an instance method?
        if 'im_class' in dir(func):
            entry_kvs['Class'] = func.im_class.__name__

        if send_entry_event:
            # log entry event
            if layer is None:
                log('profile_entry', layer, keys=entry_kvs, store_backtrace=False)
            else:
                log('entry', layer, keys=entry_kvs, store_backtrace=False)

        res = None  # return value of wrapped function
        stats = None  # cProfile statistics, if enabled
        try:
            if profile and found_cprofile:  # use cProfile?
                p = cProfile.Profile()
                res = p.runcall(func, *f_args, **f_kwargs)  # call func via cProfile
                stats = _get_profile_info(p)
            else:  # don't use cProfile, call func directly
                res = func(*f_args, **f_kwargs)
        except Exception:
            # log exception and re-raise
            log_exception()
            raise
        finally:
            # prepare data for reporting exit event
            exit_kvs = {}
            edge_str = None

            # call the callback function, if set, and merge its return
            # values with the exit event's reporting data
            if callback and callable(callback):
                try:
                    cb_ret = callback(func, f_args, f_kwargs, res)
                    # callback can optionally return a 2-tuple, where the
                    # second parameter is an additional edge to add to
                    # the exit event
                    if isinstance(cb_ret, tuple) and len(cb_ret) == 2:
                        cb_ret, edge_str = cb_ret
                    if cb_ret:
                        exit_kvs.update(cb_ret)
                except Exception:
                    # should be no user exceptions here; it's a trace-related callback
                    type_, msg_, bt_ = sys.exc_info()
                    if not TEST_MODE:
                        logger.debug("Non-fatal error in log_method callback: %s, %s, %s"
                                     % (str(type_), msg_, _str_backtrace(bt_)))
                    del bt_

            # (optionally) report return value
            if store_return:
                exit_kvs['ReturnValue'] = str(res)

            # (optionally) report profiler results
            if profile and stats:
                exit_kvs['ProfileStats'] = stats

            if send_exit_event:
                # log exit event
                if layer is None:
                    log(
                        'profile_exit', layer, keys=exit_kvs,
                        store_backtrace=False, edge_str=edge_str
                        )
                else:
                    log(
                        'exit', layer, keys=exit_kvs,
                        store_backtrace=False, edge_str=edge_str
                        )
        return res  # return output of func(*f_args, **f_kwargs)

    _log_method_wrapper._appoptics_apm_wrapped = True  # mark our wrapper for protection below

    # instrumentation helper decorator, called to add wrapper at "decorate" time
    def decorate_with_log_method(f):
        if getattr(f, '_appoptics_apm_wrapped', False):  # has this function already been wrapped?
            return f  # then pass through
        if hasattr(f, '__func__'):  # Is this a bound method of an object
            f = f.__func__  # then wrap the unbound method
        logger.debug('decorate_with_log_method wraps: {f}'.format(f= f.__name__))
        return decorator(_log_method_wrapper, f)  # otherwise wrap function f with wrapper

    # return decorator function with arguments to log_method() baked in
    return decorate_with_log_method


def _reporter():
    """A reporter has already been initialized by appoptics_apm_init() when appoptics_apm is imported.
    However, when a new child process is forked it needs its own reporter. Inside oboe_init_reporter()
    it will check if the current pid matches the stored one, if not it will create a new reporter."""

    global reporter_instance

    if not reporter_instance:
        config_str = ','.join(
            ('port=%s' % (config.get('reporter_port')),
             'cid=%s' % (config.get('service_key')))
            )
        reporter_instance = Reporter(config['collector_mode'], config_str)

    return reporter_instance


def _Event_addInfo_safe(func):
    def wrapped(*args, **kw):
        try:  # call SWIG-generated Event.addInfo (from swig/oboe.py)
            return func(*args, **kw)
        except NotImplementedError:  # unrecognized type passed to addInfo SWIG binding
            # args: [self, KeyName, Value]
            if len(args) == 3 and isinstance(args[1], six.string_types):
                # report this error
                func(args[0], '_Warning', 'Bad type for %s: %s' % (args[1], type(args[2])))
                # last resort: coerce type to string
                if hasattr(args[2], '__str__'):
                    try:
                        return func(args[0], args[1], str(args[2]))
                    except UnicodeEncodeError:
                        return func(args[0], args[1], args[2].encode('utf-8'))

                elif hasattr(args[2], '__repr__'):
                    return func(args[0], args[1], repr(args[2]))

    return wrapped


def sample_request(layer, xtr):
    """This functions calls the liboboe API to get the sampling decision. A side effect
    is to get the server-side sample_rate and sample_source.
    """
    rv = SwigContext.sampleRequest(layer, xtr or '')

    # For older binding to liboboe that returns true/false, just return that.
    if rv.__class__ == bool or (rv == 0):
        return rv

    # Newer binding to liboboe returns a bit masked integer with SampleRate and
    # Source embedded
    config['sample_rate'] = ((rv & SAMPLE_RATE_MASK) / 1e6)
    config['sample_source'] = (rv & SAMPLE_SOURCE_MASK) >> 24
    setattr(Context.transaction_dict, 'sample_rate', config['sample_rate'])
    setattr(Context.transaction_dict, 'sample_source', config['sample_source'])

    return rv


###############################################################################
# Backwards compatability
###############################################################################

setattr(SwigEvent, 'addInfo', _Event_addInfo_safe(getattr(SwigEvent, 'addInfo')))


def _old_context_log(cls, layer, label, backtrace=False, **kwargs):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.log is deprecated. '
            'Please use appoptics_apm.log (and note signature change).')
    log(label, layer, store_backtrace=backtrace, keys=kwargs)


def _old_context_log_error(cls, exception=None, err_class=None, err_msg=None, backtrace=True):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.log_error is deprecated. '
            'Please use appoptics_apm.log_error (and note signature change).'
        )
    if exception:
        err_class = exception.__class__.__name__
        err_msg = str(exception)
    store_backtrace = False
    tb = None
    if backtrace:
        _, _, tb = sys.exc_info()
        store_backtrace = True
    try:
        return log_error(err_class, err_msg, store_backtrace=store_backtrace, backtrace=tb)
    finally:
        del tb


def _old_context_log_exception(cls, msg=None, exc_info=None, backtrace=True):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.log_exception is deprecated. '
            'Please use appoptics_apm.log_exception (and note signature change).'
        )
    typ, val, tb = exc_info or sys.exc_info()
    if msg is None:
        try:
            msg = str(val)
        except Exception:
            msg = repr(val)
    try:
        return log_error(typ.__name__, msg, store_backtrace=backtrace, backtrace=tb)
    finally:
        del tb


def _old_context_trace(cls, layer='Python', xtr_hdr=None, kvs=None):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.trace is deprecated. '
            'Please use appoptics_apm.trace (and note signature change).'
        )
    return trace(layer, xtr_hdr=xtr_hdr, kvs=kvs)


def _old_context_profile_function(cls, profile_name, store_args=False, store_return=False,
    store_backtrace=False, profile=False, callback=None, **entry_kvs):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.trace is deprecated. '
            'Please use appoptics_apm.trace (and note signature change).'
        )
    return profile_function(
        profile_name, store_args=False, store_return=False, store_backtrace=False,
        profile=False, callback=None, entry_kvs=entry_kvs
    )


def _old_context_log_method(cls, layer='Python', store_return=False, store_args=False,
                            callback=None, profile=False, **entry_kvs):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.log_method is deprecated. '
            'Please use appoptics_apm.log_method (and note signature change).'
        )
    return log_method(layer, store_return=store_return, store_args=store_args,
                      callback=callback, profile=profile, entry_kvs=entry_kvs)


class _old_context_profile_block(profile_block):
    def __init__(self, *args, **kw):
        if config['warn_deprecated']:
            logger.debug(
                'appoptics_apm.Context.profile_block is deprecated. '
                'Please use appoptics_apm.profile_block (and note signature change).'
            )
        super(_old_context_profile_block, self).__init__(*args, **kw)


def _old_context_to_string(cls):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.toString is deprecated. '
            'Please use str(appoptics_apm.Context.get_default())'
        )
    return str(Context.get_default())


def _old_context_from_string(cls, md_string):
    if config['warn_deprecated']:
        logger.debug('appoptics_apm.Context.fromString is deprecated.')
    c = Context(md_string)
    c.set_as_default()


def _old_context_is_valid(cls):
    if config['warn_deprecated']:
        logger.debug(
            'appoptics_apm.Context.isValid is deprecated. '
            'Please use appoptics_apm.Context.get_default().is_valid()'
        )
    return Context.get_default().is_valid()


setattr(Context, 'log', types.MethodType(_old_context_log, Context))
setattr(Context, 'log_error', types.MethodType(_old_context_log_error, Context))
setattr(Context, 'log_exception', types.MethodType(_old_context_log_exception, Context))
setattr(Context, 'log_method', types.MethodType(_old_context_log_method, Context))
setattr(Context, 'trace', types.MethodType(_old_context_trace, Context))
setattr(Context, 'profile_function', types.MethodType(_old_context_profile_function, Context))
setattr(Context, 'profile_block', _old_context_profile_block)
setattr(Context, 'toString', types.MethodType(_old_context_to_string, Context))
setattr(Context, 'fromString', types.MethodType(_old_context_from_string, Context))
setattr(Context, 'isValid', types.MethodType(_old_context_is_valid, Context))


def report_layer_init(layer='Python', keys=None):
    """ Send a status report with postStatus Thrift API showing the initialization and version of
    this layer's instrumentation.
    """
    if not ready():
        logger.debug('AppOptics APM is not ready, ignoring init message of layer {layer}.'.format(layer=layer))
        return

    ver_keys = dict()
    keys = keys or dict()

    ver_keys['__Init'] = 'True'
    ver_keys['Python.Version'] = sys.version
    ver_keys['Python.AppOptics.Version'] = __version__

    ver_keys['Python.InstallDirectory'] = os.path.dirname(__file__)
    ver_keys['Python.InstallTimestamp'] = os.path.getmtime(__file__)  # in sec since epoch
    ver_keys['Python.LastRestart'] = AGENT_START_TIME  # in usec

    # Don't add Hostname here as liboboe will do it for you.
    # ver_keys['Hostname'] = socket.gethostname()

    if layer.lower() == 'tornado':
        try:
            import tornado
            ver_keys["Python.Tornado.Version"] = sys.modules['tornado'].version
        except ImportError as e:
            logger.warning('Failed to report init event for Tornado: {e}'.format(e=e))
            return

    if layer.lower() == 'django':
        try:
            import django
            ver_keys["Python.Django.Version"] = django.get_version()
        except ImportError as e:
            logger.warning('Failed to report init event for Django: {e}'.format(e=e))
            return

    ver_keys.update(keys)

    ctx = Context(Metadata.makeRandom(True))
    if not ctx.is_valid():
        return
    evt = ctx.create_event('single', layer)

    for k, v in ver_keys.items():
        evt.add_info(k, v)
    ctx.report_status(evt)

# Report an status event after everything is done.
report_layer_init('Python')

class oboe_ready_code():
    OBOE_SERVER_RESPONSE_UNKNOWN = (0,"Oboe server : unknown error")
    OBOE_SERVER_RESPONSE_OK = (1,"Oboe server : is ready")
    OBOE_SERVER_RESPONSE_TRY_LATER = (2,"Oboe server : not ready yet, try later")
    OBOE_SERVER_RESPONSE_LIMIT_EXCEEDED = (3,"Oboe server : limit exceeded")
    OBOE_SERVER_RESPONSE_INVALID_API_KEY = (4,"Oboe server : invalid API key")
    OBOE_SERVER_RESPONSE_CONNECT_ERROR = (5,"Oboe server : connection error")

    @classmethod
    def code_values(cls):
        code_pairs = [v for k,v in cls.__dict__.items() if not k.startswith("__")]
        return {p[0]:p[1] for p in code_pairs if isinstance(p, tuple)}

def appoptics_ready(wait_milliseconds = 3000, integer_response = False):
    """
     Wait for AppOptics to be ready to send traces.
    
     This may be useful in short lived background processes when it is important to capture
     information during the whole time the process is running. Usually AppOptics doesn't block an
     application while it is starting up.
    
     :param wait_milliseconds:int default 3000, the maximum time to wait in milliseconds
     :param integer_response:int default false, return boolean value, otherwise return integer for
     detail information 

     :return: return True for ready, False not ready, integer 1 for ready, others not ready 
    
     :Example:
    
      if not appoptics_ready(10000):
         Logger.info("AppOptics not ready after 10 seconds, no metrics will be sent")
    """

    rc = SwigContext.isReady(wait_milliseconds)
    if not type(rc) is int or not rc in oboe_ready_code.code_values():
        logger.warning("Unrecognized  return code: " + str(rc))
    elif rc != oboe_ready_code.OBOE_SERVER_RESPONSE_OK[0]:
        logger.warning(oboe_ready_code.code_values()[rc])

    return  rc if integer_response else rc == oboe_ready_code.OBOE_SERVER_RESPONSE_OK[0]



""" AppOptics APM instrumentation for Django's template system

Copyright (C) 2016 SolarWinds, LLC.
All rights reserved.
"""
from appoptics_apm import util

appoptics_apm_logger = util.logger

TEMPLATE_LAYER = 'template'


def before_render_cb(func, f_args, f_kwargs):
    """ Callback to django.template.base.Template.render(),
        retrieves template name from Template object and
        manually turns a log_method into a profile_function,
        (necessary to dynamically generate profile name...) """
    template_name = f_args[0].name
    keys = {
        'Language': 'template',
        'ProfileName': template_name
    }
    return f_args, f_kwargs, keys


def after_render_cb(func, f_args, f_kwargs, ret):
    template_name = f_args[0].name
    keys = {
        'Language': 'template',
        'ProfileName': template_name
    }
    return keys, None


def wrap(module):
    try:
        # profile on specific template name (inner wrap)
        profile_wrapper_render = util.log_method(
            None, before_callback=before_render_cb,
            callback=after_render_cb
        )

        # overall template layer (outer wrap)
        layer_wrapper_render = util.log_method(TEMPLATE_LAYER)

        setattr(module.Template, 'render',
                layer_wrapper_render(profile_wrapper_render(module.Template.render)))

    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM error: %s" % str(e))
